@@ -2,26 +2,19 @@
 
   error.c -
 
-  $Author: knu $
-  $Date: 2008-05-31 22:37:06 +0900 (Sat, 31 May 2008) $
+  $Author: nobu $
   created at: Mon Aug  9 16:11:34 JST 1993
 
-  Copyright (C) 1993-2003 Yukihiro Matsumoto
+  Copyright (C) 1993-2007 Yukihiro Matsumoto
 
 **********************************************************************/
 
-#include "ruby.h"
-#include "env.h"
-#include "st.h"
+#include "ruby/ruby.h"
+#include "ruby/st.h"
+#include "vm_core.h"
 
 #include <stdio.h>
-#ifdef HAVE_STDARG_PROTOTYPES
 #include <stdarg.h>
-#define va_init_list(a,b) va_start(a,b)
-#else
-#include <varargs.h>
-#define va_init_list(a,b) va_start(a)
-#endif
 #ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
@@ -29,33 +22,30 @@
 #define EXIT_SUCCESS 0
 #endif
 
-extern const char *ruby_description;
-
-int ruby_nerrs;
+extern const char ruby_description[];
 
 static int
-err_position(buf, len)
-    char *buf;
-    long len;
+err_position_0(char *buf, long len, const char *file, int line)
 {
-    ruby_set_current_source();
-    if (!ruby_sourcefile) {
+    if (!file) {
 	return 0;
     }
-    else if (ruby_sourceline == 0) {
-	return snprintf(buf, len, "%s: ", ruby_sourcefile);
+    else if (line == 0) {
+	return snprintf(buf, len, "%s: ", file);
     }
     else {
-	return snprintf(buf, len, "%s:%d: ", ruby_sourcefile, ruby_sourceline);
+	return snprintf(buf, len, "%s:%d: ", file, line);
     }
 }
 
+static int
+err_position(char *buf, long len)
+{
+    return err_position_0(buf, len, rb_sourcefile(), rb_sourceline());
+}
+
 static void
-err_snprintf(buf, len, fmt, args)
-    char *buf;
-    long len;
-    const char *fmt;
-    va_list args;
+err_snprintf(char *buf, long len, const char *fmt, va_list args)
 {
     long n;
 
@@ -65,57 +55,88 @@ err_snprintf(buf, len, fmt, args)
     }
 }
 
-static void err_append _((const char*));
 static void
-err_print(fmt, args)
-    const char *fmt;
-    va_list args;
+compile_snprintf(char *buf, long len, const char *file, int line, const char *fmt, va_list args)
 {
+    long n;
+
+    n = err_position_0(buf, len, file, line);
+    if (len > n) {
+	vsnprintf((char*)buf+n, len-n, fmt, args);
+    }
+}
+
+static void err_append(const char*);
+
+void
+rb_compile_error(const char *file, int line, const char *fmt, ...)
+{
+    va_list args;
     char buf[BUFSIZ];
 
-    err_snprintf(buf, BUFSIZ, fmt, args);
+    va_start(args, fmt);
+    compile_snprintf(buf, BUFSIZ, file, line, fmt, args);
+    va_end(args);
     err_append(buf);
 }
 
 void
-#ifdef HAVE_STDARG_PROTOTYPES
-rb_compile_error(const char *fmt, ...)
-#else
-rb_compile_error(fmt, va_alist)
-    const char *fmt;
-    va_dcl
-#endif
-{
-    va_list args;
-
-    va_init_list(args, fmt);
-    err_print(fmt, args);
-    va_end(args);
-    ruby_nerrs++;
-}
-
-void
-#ifdef HAVE_STDARG_PROTOTYPES
 rb_compile_error_append(const char *fmt, ...)
-#else
-rb_compile_error_append(fmt, va_alist)
-    const char *fmt;
-    va_dcl
-#endif
 {
     va_list args;
     char buf[BUFSIZ];
 
-    va_init_list(args, fmt);
+    va_start(args, fmt);
     vsnprintf(buf, BUFSIZ, fmt, args);
     va_end(args);
     err_append(buf);
 }
 
 static void
-warn_print(fmt, args)
-    const char *fmt;
+compile_warn_print(const char *file, int line, const char *fmt, va_list args)
+{
+    char buf[BUFSIZ];
+    int len;
+
+    compile_snprintf(buf, BUFSIZ, file, line, fmt, args);
+    len = strlen(buf);
+    buf[len++] = '\n';
+    rb_write_error2(buf, len);
+}
+
+void
+rb_compile_warn(const char *file, int line, const char *fmt, ...)
+{
+    char buf[BUFSIZ];
     va_list args;
+
+    if (NIL_P(ruby_verbose)) return;
+
+    snprintf(buf, BUFSIZ, "warning: %s", fmt);
+
+    va_start(args, fmt);
+    compile_warn_print(file, line, buf, args);
+    va_end(args);
+}
+
+/* rb_compile_warning() reports only in verbose mode */
+void
+rb_compile_warning(const char *file, int line, const char *fmt, ...)
+{
+    char buf[BUFSIZ];
+    va_list args;
+
+    if (!RTEST(ruby_verbose)) return;
+
+    snprintf(buf, BUFSIZ, "warning: %s", fmt);
+
+    va_start(args, fmt);
+    compile_warn_print(file, line, buf, args);
+    va_end(args);
+}
+
+static void
+warn_print(const char *fmt, va_list args)
 {
     char buf[BUFSIZ];
     int len;
@@ -127,13 +148,7 @@ warn_print(fmt, args)
 }
 
 void
-#ifdef HAVE_STDARG_PROTOTYPES
 rb_warn(const char *fmt, ...)
-#else
-rb_warn(fmt, va_alist)
-    const char *fmt;
-    va_dcl
-#endif
 {
     char buf[BUFSIZ];
     va_list args;
@@ -142,20 +157,14 @@ rb_warn(fmt, va_alist)
 
     snprintf(buf, BUFSIZ, "warning: %s", fmt);
 
-    va_init_list(args, fmt);
+    va_start(args, fmt);
     warn_print(buf, args);
     va_end(args);
 }
 
 /* rb_warning() reports only in verbose mode */
 void
-#ifdef HAVE_STDARG_PROTOTYPES
 rb_warning(const char *fmt, ...)
-#else
-rb_warning(fmt, va_alist)
-    const char *fmt;
-    va_dcl
-#endif
 {
     char buf[BUFSIZ];
     va_list args;
@@ -164,7 +173,7 @@ rb_warning(fmt, va_alist)
 
     snprintf(buf, BUFSIZ, "warning: %s", fmt);
 
-    va_init_list(args, fmt);
+    va_start(args, fmt);
     warn_print(buf, args);
     va_end(args);
 }
@@ -178,8 +187,7 @@ rb_warning(fmt, va_alist)
  */
 
 static VALUE
-rb_warn_m(self, mesg)
-    VALUE self, mesg;
+rb_warn_m(VALUE self, VALUE mesg)
 {
     if (!NIL_P(ruby_verbose)) {
 	rb_io_write(rb_stderr, mesg);
@@ -188,32 +196,49 @@ rb_warn_m(self, mesg)
     return Qnil;
 }
 
-void
-#ifdef HAVE_STDARG_PROTOTYPES
-rb_bug(const char *fmt, ...)
-#else
-rb_bug(fmt, va_alist)
-    const char *fmt;
-    va_dcl
-#endif
+void rb_vm_bugreport(void);
+
+static void
+report_bug(const char *file, int line, const char *fmt, va_list args)
 {
     char buf[BUFSIZ];
-    va_list args;
     FILE *out = stderr;
-    int len = err_position(buf, BUFSIZ);
+    int len = err_position_0(buf, BUFSIZ, file, line);
 
     if (fwrite(buf, 1, len, out) == len ||
 	fwrite(buf, 1, len, (out = stdout)) == len) {
 	fputs("[BUG] ", out);
-	va_init_list(args, fmt);
 	vfprintf(out, fmt, args);
-	va_end(args);
 	fprintf(out, "\n%s\n\n", ruby_description);
+	rb_vm_bugreport();
     }
+}
+
+void
+rb_bug(const char *fmt, ...)
+{
+    va_list args;
+
+    va_start(args, fmt);
+    report_bug(rb_sourcefile(), rb_sourceline(), fmt, args);
+    va_end(args);
+
     abort();
 }
 
-static struct types {
+void
+rb_compile_bug(const char *file, int line, const char *fmt, ...)
+{
+    va_list args;
+
+    va_start(args, fmt);
+    report_bug(file, line, fmt, args);
+    va_end(args);
+
+    abort();
+}
+
+static const struct types {
     int type;
     const char *name;
 } builtin_types[] = {
@@ -231,31 +256,30 @@ static struct types {
     {T_STRUCT,	"Struct"},
     {T_BIGNUM,	"Bignum"},
     {T_FILE,	"File"},
+    {T_RATIONAL,"Rational"},
+    {T_COMPLEX, "Complex"},
     {T_TRUE,	"true"},
     {T_FALSE,	"false"},
     {T_SYMBOL,	"Symbol"},	/* :symbol */
     {T_DATA,	"Data"},	/* internal use: wrapped C pointers */
     {T_MATCH,	"MatchData"},	/* data of $~ */
-    {T_VARMAP,	"Varmap"},	/* internal use: dynamic variables */
-    {T_SCOPE,	"Scope"},	/* internal use: variable scope */
     {T_NODE,	"Node"},	/* internal use: syntax tree node */
     {T_UNDEF,	"undef"},	/* internal use: #undef; should not happen */
-    {-1,	0}
 };
 
 void
-rb_check_type(x, t)
-    VALUE x;
-    int t;
+rb_check_type(VALUE x, int t)
 {
-    struct types *type = builtin_types;
+    const struct types *type = builtin_types;
+    const struct types *const typeend = builtin_types +
+	sizeof(builtin_types) / sizeof(builtin_types[0]);
 
     if (x == Qundef) {
 	rb_bug("undef leaked to the Ruby space");
     }
 
     if (TYPE(x) != t) {
-	while (type->type >= 0) {
+	while (type < typeend) {
 	    if (type->type == t) {
 		const char *etype;
 
@@ -269,7 +293,7 @@ rb_check_type(x, t)
 		    etype = "Symbol";
 		}
 		else if (rb_special_const_p(x)) {
-		    etype = RSTRING(rb_obj_as_string(x))->ptr;
+		    etype = RSTRING_PTR(rb_obj_as_string(x));
 		}
 		else {
 		    etype = rb_obj_classname(x);
@@ -279,7 +303,7 @@ rb_check_type(x, t)
 	    }
 	    type++;
 	}
-	rb_bug("unknown type 0x%x", t);
+	rb_bug("unknown type 0x%x (0x%x given)", t, TYPE(x));
     }
 }
 
@@ -296,6 +320,7 @@ VALUE rb_eRuntimeError;
 VALUE rb_eTypeError;
 VALUE rb_eArgError;
 VALUE rb_eIndexError;
+VALUE rb_eKeyError;
 VALUE rb_eRangeError;
 VALUE rb_eNameError;
 VALUE rb_eNoMethodError;
@@ -310,30 +335,25 @@ VALUE rb_eLoadError;
 
 VALUE rb_eSystemCallError;
 VALUE rb_mErrno;
+static VALUE rb_eNOERROR;
 
 VALUE
-rb_exc_new(etype, ptr, len)
-    VALUE etype;
-    const char *ptr;
-    long len;
+rb_exc_new(VALUE etype, const char *ptr, long len)
 {
     return rb_funcall(etype, rb_intern("new"), 1, rb_str_new(ptr, len));
 }
 
 VALUE
-rb_exc_new2(etype, s)
-    VALUE etype;
-    const char *s;
+rb_exc_new2(VALUE etype, const char *s)
 {
     return rb_exc_new(etype, s, strlen(s));
 }
 
 VALUE
-rb_exc_new3(etype, str)
-    VALUE etype, str;
+rb_exc_new3(VALUE etype, VALUE str)
 {
     StringValue(str);
-    return rb_exc_new(etype, RSTRING(str)->ptr, RSTRING(str)->len);
+    return rb_funcall(etype, rb_intern("new"), 1, str);
 }
 
 /*
@@ -345,10 +365,7 @@ rb_exc_new3(etype, str)
  */
 
 static VALUE
-exc_initialize(argc, argv, exc)
-    int argc;
-    VALUE *argv;
-    VALUE exc;
+exc_initialize(int argc, VALUE *argv, VALUE exc)
 {
     VALUE arg;
 
@@ -373,10 +390,7 @@ exc_initialize(argc, argv, exc)
  */
 
 static VALUE
-exc_exception(argc, argv, self)
-    int argc;
-    VALUE *argv;
-    VALUE self;
+exc_exception(int argc, VALUE *argv, VALUE self)
 {
     VALUE exc;
 
@@ -397,8 +411,7 @@ exc_exception(argc, argv, self)
  */
 
 static VALUE
-exc_to_s(exc)
-    VALUE exc;
+exc_to_s(VALUE exc)
 {
     VALUE mesg = rb_attr_get(exc, rb_intern("mesg"));
 
@@ -410,7 +423,6 @@ exc_to_s(exc)
 /*
  * call-seq:
  *   exception.message   =>  string
- *   exception.to_str    =>  string
  *
  * Returns the result of invoking <code>exception.to_s</code>.
  * Normally this returns the exception's message or name. By
@@ -419,8 +431,7 @@ exc_to_s(exc)
  */
 
 static VALUE
-exc_to_str(exc)
-    VALUE exc;
+exc_message(VALUE exc)
 {
     return rb_funcall(exc, rb_intern("to_s"), 0, 0);
 }
@@ -433,14 +444,13 @@ exc_to_str(exc)
  */
 
 static VALUE
-exc_inspect(exc)
-    VALUE exc;
+exc_inspect(VALUE exc)
 {
     VALUE str, klass;
 
     klass = CLASS_OF(exc);
     exc = rb_obj_as_string(exc);
-    if (RSTRING(exc)->len == 0) {
+    if (RSTRING_LEN(exc) == 0) {
 	return rb_str_dup(rb_class_name(klass));
     }
 
@@ -484,18 +494,16 @@ exc_inspect(exc)
 */
 
 static VALUE
-exc_backtrace(exc)
-    VALUE exc;
+exc_backtrace(VALUE exc)
 {
-    static ID bt;
+    ID bt;
 
-    if (!bt) bt = rb_intern("bt");
+    CONST_ID(bt, "bt");
     return rb_attr_get(exc, bt);
 }
 
 VALUE
-rb_check_backtrace(bt)
-    VALUE bt;
+rb_check_backtrace(VALUE bt)
 {
     long i;
     static const char err[] = "backtrace must be Array of String";
@@ -507,8 +515,8 @@ rb_check_backtrace(bt)
 	if (t != T_ARRAY) {
 	    rb_raise(rb_eTypeError, err);
 	}
-	for (i=0;i<RARRAY(bt)->len;i++) {
-	    if (TYPE(RARRAY(bt)->ptr[i]) != T_STRING) {
+	for (i=0;i<RARRAY_LEN(bt);i++) {
+	    if (TYPE(RARRAY_PTR(bt)[i]) != T_STRING) {
 		rb_raise(rb_eTypeError, err);
 	    }
 	}
@@ -527,13 +535,36 @@ rb_check_backtrace(bt)
  */
 
 static VALUE
-exc_set_backtrace(exc, bt)
-    VALUE exc;
-    VALUE bt;
+exc_set_backtrace(VALUE exc, VALUE bt)
 {
     return rb_iv_set(exc, "bt", rb_check_backtrace(bt));
 }
 
+/*
+ *  call-seq:
+ *     exc == obj   => true or false
+ *  
+ *  Equality---If <i>obj</i> is not an <code>Exception</code>, returns
+ *  <code>false</code>. Otherwise, returns <code>true</code> if <i>exc</i> and 
+ *  <i>obj</i> share same class, messages, and backtrace.
+ */
+
+static VALUE
+exc_equal(VALUE exc, VALUE obj)
+{
+    ID id_mesg;
+
+    if (exc == obj) return Qtrue;
+    if (rb_obj_class(exc) != rb_obj_class(obj))
+	return rb_equal(obj, exc);
+    CONST_ID(id_mesg, "mesg");
+    if (!rb_equal(rb_attr_get(exc, id_mesg), rb_attr_get(obj, id_mesg)))
+	return Qfalse;
+    if (!rb_equal(exc_backtrace(exc), exc_backtrace(obj)))
+	return Qfalse;
+    return Qtrue;
+}
+
 /*
  * call-seq:
  *   SystemExit.new(status=0)   => system_exit
@@ -542,10 +573,7 @@ exc_set_backtrace(exc, bt)
  */
 
 static VALUE
-exit_initialize(argc, argv, exc)
-    int argc;
-    VALUE *argv;
-    VALUE exc;
+exit_initialize(int argc, VALUE *argv, VALUE exc)
 {
     VALUE status = INT2FIX(EXIT_SUCCESS);
     if (argc > 0 && FIXNUM_P(argv[0])) {
@@ -566,8 +594,7 @@ exit_initialize(argc, argv, exc)
  */
 
 static VALUE
-exit_status(exc)
-    VALUE exc;
+exit_status(VALUE exc)
 {
     return rb_attr_get(exc, rb_intern("status"));
 }
@@ -581,8 +608,7 @@ exit_status(exc)
  */
 
 static VALUE
-exit_success_p(exc)
-    VALUE exc;
+exit_success_p(VALUE exc)
 {
     VALUE status = rb_attr_get(exc, rb_intern("status"));
     if (NIL_P(status)) return Qtrue;
@@ -591,24 +617,15 @@ exit_success_p(exc)
 }
 
 void
-#ifdef HAVE_STDARG_PROTOTYPES
 rb_name_error(ID id, const char *fmt, ...)
-#else
-rb_name_error(id, fmt, va_alist)
-    ID id;
-    const char *fmt;
-    va_dcl
-#endif
 {
     VALUE exc, argv[2];
     va_list args;
-    char buf[BUFSIZ];
 
-    va_init_list(args, fmt);
-    vsnprintf(buf, BUFSIZ, fmt, args);
+    va_start(args, fmt);
+    argv[0] = rb_vsprintf(fmt, args);
     va_end(args);
 
-    argv[0] = rb_str_new2(buf);
     argv[1] = ID2SYM(id);
     exc = rb_class_new_instance(2, argv, rb_eNameError);
     rb_exc_raise(exc);
@@ -624,10 +641,7 @@ rb_name_error(id, fmt, va_alist)
  */
 
 static VALUE
-name_err_initialize(argc, argv, self)
-    int argc;
-    VALUE *argv;
-    VALUE self;
+name_err_initialize(int argc, VALUE *argv, VALUE self)
 {
     VALUE name;
 
@@ -645,8 +659,7 @@ name_err_initialize(argc, argv, self)
  */
 
 static VALUE
-name_err_name(self)
-    VALUE self;
+name_err_name(VALUE self)
 {
     return rb_attr_get(self, rb_intern("name"));
 }
@@ -655,14 +668,14 @@ name_err_name(self)
  * call-seq:
  *  name_error.to_s   => string
  *
- * Produce a nicely-formated string representing the +NameError+.
+ * Produce a nicely-formatted string representing the +NameError+.
  */
 
 static VALUE
-name_err_to_s(exc)
-    VALUE exc;
+name_err_to_s(VALUE exc)
 {
-    VALUE mesg = rb_attr_get(exc, rb_intern("mesg")), str = mesg;
+    VALUE mesg = rb_attr_get(exc, rb_intern("mesg"));
+    VALUE str = mesg;
 
     if (NIL_P(mesg)) return rb_class_name(CLASS_OF(exc));
     StringValue(str);
@@ -684,10 +697,7 @@ name_err_to_s(exc)
  */
 
 static VALUE
-nometh_err_initialize(argc, argv, self)
-    int argc;
-    VALUE *argv;
-    VALUE self;
+nometh_err_initialize(int argc, VALUE *argv, VALUE self)
 {
     VALUE args = (argc > 2) ? argv[--argc] : Qnil;
     name_err_initialize(argc, argv, self);
@@ -697,16 +707,14 @@ nometh_err_initialize(argc, argv, self)
 
 /* :nodoc: */
 static void
-name_err_mesg_mark(ptr)
-    VALUE *ptr;
+name_err_mesg_mark(VALUE *ptr)
 {
     rb_gc_mark_locations(ptr, ptr+3);
 }
 
 /* :nodoc: */
 static VALUE
-name_err_mesg_new(obj, mesg, recv, method)
-    VALUE obj, mesg, recv, method;
+name_err_mesg_new(VALUE obj, VALUE mesg, VALUE recv, VALUE method)
 {
     VALUE *ptr = ALLOC_N(VALUE, 3);
 
@@ -718,8 +726,27 @@ name_err_mesg_new(obj, mesg, recv, method)
 
 /* :nodoc: */
 static VALUE
-name_err_mesg_to_str(obj)
-    VALUE obj;
+name_err_mesg_equal(VALUE obj1, VALUE obj2)
+{
+    VALUE *ptr1, *ptr2;
+    int i;
+
+    if (obj1 == obj2) return Qtrue;
+    if (rb_obj_class(obj2) != rb_cNameErrorMesg)
+	return Qfalse;
+
+    Data_Get_Struct(obj1, VALUE, ptr1);
+    Data_Get_Struct(obj2, VALUE, ptr2);
+    for (i=0; i<3; i++) {
+	if (!rb_equal(ptr1[i], ptr2[i]))
+	    return Qfalse;
+    }
+    return Qtrue;
+}
+
+/* :nodoc: */
+static VALUE
+name_err_mesg_to_str(VALUE obj)
 {
     VALUE *ptr, mesg;
     Data_Get_Struct(obj, VALUE, ptr);
@@ -743,10 +770,10 @@ name_err_mesg_to_str(obj)
 	    break;
 	  default:
 	    d = rb_protect(rb_inspect, obj, 0);
-	    if (NIL_P(d) || RSTRING(d)->len > 65) {
+	    if (NIL_P(d) || RSTRING_LEN(d) > 65) {
 		d = rb_any_to_s(obj);
 	    }
-	    desc = RSTRING(d)->ptr;
+	    desc = RSTRING_PTR(d);
 	    break;
 	}
 	if (desc && desc[0] != '#') {
@@ -765,8 +792,7 @@ name_err_mesg_to_str(obj)
 
 /* :nodoc: */
 static VALUE
-name_err_mesg_load(klass, str)
-    VALUE klass, str;
+name_err_mesg_load(VALUE klass, VALUE str)
 {
     return str;
 }
@@ -780,19 +806,17 @@ name_err_mesg_load(klass, str)
  */
 
 static VALUE
-nometh_err_args(self)
-    VALUE self;
+nometh_err_args(VALUE self)
 {
     return rb_attr_get(self, rb_intern("args"));
 }
 
 void
-rb_invalid_str(str, type)
-    const char *str, *type;
+rb_invalid_str(const char *str, const char *type)
 {
     VALUE s = rb_str_inspect(rb_str_new2(str));
 
-    rb_raise(rb_eArgError, "invalid value for %s: %s", type, RSTRING(s)->ptr);
+    rb_raise(rb_eArgError, "invalid value for %s: %s", type, RSTRING_PTR(s));
 }
 
 /* 
@@ -823,15 +847,13 @@ rb_invalid_str(str, type)
  *  The full list of operating system errors on your particular platform
  *  are available as the constants of <code>Errno</code>.
  *
- *     Errno.constants   #=> E2BIG, EACCES, EADDRINUSE, EADDRNOTAVAIL, ...
+ *     Errno.constants   #=> :E2BIG, :EACCES, :EADDRINUSE, :EADDRNOTAVAIL, ...
  */
 
 static st_table *syserr_tbl;
 
 static VALUE
-set_syserr(n, name)
-    int n;
-    const char *name;
+set_syserr(int n, const char *name)
 {
     VALUE error;
 
@@ -847,8 +869,7 @@ set_syserr(n, name)
 }
 
 static VALUE
-get_syserr(n)
-    int n;
+get_syserr(int n)
 {
     VALUE error;
 
@@ -873,10 +894,7 @@ get_syserr(n)
  */
 
 static VALUE
-syserr_initialize(argc, argv, self)
-    int argc;
-    VALUE *argv;
-    VALUE self;
+syserr_initialize(int argc, VALUE *argv, VALUE self)
 {
 #if !defined(_WIN32) && !defined(__VMS)
     char *strerror();
@@ -906,14 +924,10 @@ syserr_initialize(argc, argv, self)
     else err = "unknown error";
     if (!NIL_P(mesg)) {
 	VALUE str = mesg;
-	size_t len;
 
 	StringValue(str);
-	len = strlen(err)+RSTRING(str)->len+3;
-	mesg = rb_str_new(0, len);
-	snprintf(RSTRING(mesg)->ptr, len+1, "%s - %.*s", err,
-		(int)RSTRING(str)->len, RSTRING(str)->ptr);
-	rb_str_resize(mesg, strlen(RSTRING(mesg)->ptr));
+	mesg = rb_sprintf("%s - %.*s", err,
+			  (int)RSTRING_LEN(str), RSTRING_PTR(str));
     }
     else {
 	mesg = rb_str_new2(err);
@@ -931,8 +945,7 @@ syserr_initialize(argc, argv, self)
  */
 
 static VALUE
-syserr_errno(self)
-    VALUE self;
+syserr_errno(VALUE self)
 {
     return rb_attr_get(self, rb_intern("errno"));
 }
@@ -945,20 +958,20 @@ syserr_errno(self)
  * if the error numbers _self_ and _other_ are the same.
  */
 
-
 static VALUE
-syserr_eqq(self, exc)
-    VALUE self, exc;
+syserr_eqq(VALUE self, VALUE exc)
 {
     VALUE num, e;
-    ID en = rb_intern("errno");
+    ID en;
+
+    CONST_ID(en, "errno");
 
     if (!rb_obj_is_kind_of(exc, rb_eSystemCallError)) {
 	if (!rb_respond_to(exc, en)) return Qfalse;
     }
     else if (self == rb_eSystemCallError) return Qtrue;
 
-    num = rb_attr_get(exc, en);
+    num = rb_attr_get(exc, rb_intern("errno"));
     if (NIL_P(num)) {
 	num = rb_funcall(exc, en, 0, 0);
     }
@@ -969,7 +982,19 @@ syserr_eqq(self, exc)
 }
 
 /*
- *  Descendents of class <code>Exception</code> are used to communicate
+ * call-seq:
+ *   Errno.const_missing   => SystemCallError
+ *
+ * Returns default SystemCallError class.
+ */
+static VALUE
+errno_missing(VALUE self, VALUE id)
+{
+    return rb_eNOERROR;
+}
+
+/*
+ *  Descendants of class <code>Exception</code> are used to communicate
  *  between <code>raise</code> methods and <code>rescue</code>
  *  statements in <code>begin/end</code> blocks. <code>Exception</code>
  *  objects carry information about the exception---its type (the
@@ -979,15 +1004,15 @@ syserr_eqq(self, exc)
  */
 
 void
-Init_Exception()
+Init_Exception(void)
 {
     rb_eException   = rb_define_class("Exception", rb_cObject);
     rb_define_singleton_method(rb_eException, "exception", rb_class_new_instance, -1);
     rb_define_method(rb_eException, "exception", exc_exception, -1);
     rb_define_method(rb_eException, "initialize", exc_initialize, -1);
+    rb_define_method(rb_eException, "==", exc_equal, 1);
     rb_define_method(rb_eException, "to_s", exc_to_s, 0);
-    rb_define_method(rb_eException, "to_str", exc_to_str, 0);
-    rb_define_method(rb_eException, "message", exc_to_str, 0);
+    rb_define_method(rb_eException, "message", exc_message, 0);
     rb_define_method(rb_eException, "inspect", exc_inspect, 0);
     rb_define_method(rb_eException, "backtrace", exc_backtrace, 0);
     rb_define_method(rb_eException, "set_backtrace", exc_set_backtrace, 1);
@@ -1005,13 +1030,21 @@ Init_Exception()
     rb_eTypeError     = rb_define_class("TypeError", rb_eStandardError);
     rb_eArgError      = rb_define_class("ArgumentError", rb_eStandardError);
     rb_eIndexError    = rb_define_class("IndexError", rb_eStandardError);
+    rb_eKeyError      = rb_define_class("KeyError", rb_eIndexError);
     rb_eRangeError    = rb_define_class("RangeError", rb_eStandardError);
+
+    rb_eScriptError = rb_define_class("ScriptError", rb_eException);
+    rb_eSyntaxError = rb_define_class("SyntaxError", rb_eScriptError);
+    rb_eLoadError   = rb_define_class("LoadError", rb_eScriptError);
+    rb_eNotImpError = rb_define_class("NotImplementedError", rb_eScriptError);
+
     rb_eNameError     = rb_define_class("NameError", rb_eStandardError);
     rb_define_method(rb_eNameError, "initialize", name_err_initialize, -1);
     rb_define_method(rb_eNameError, "name", name_err_name, 0);
     rb_define_method(rb_eNameError, "to_s", name_err_to_s, 0);
     rb_cNameErrorMesg = rb_define_class_under(rb_eNameError, "message", rb_cData);
     rb_define_singleton_method(rb_cNameErrorMesg, "!", name_err_mesg_new, 3);
+    rb_define_method(rb_cNameErrorMesg, "==", name_err_mesg_equal, 1);
     rb_define_method(rb_cNameErrorMesg, "to_str", name_err_mesg_to_str, 0);
     rb_define_method(rb_cNameErrorMesg, "_dump", name_err_mesg_to_str, 1);
     rb_define_singleton_method(rb_cNameErrorMesg, "_load", name_err_mesg_load, 1);
@@ -1019,13 +1052,8 @@ Init_Exception()
     rb_define_method(rb_eNoMethodError, "initialize", nometh_err_initialize, -1);
     rb_define_method(rb_eNoMethodError, "args", nometh_err_args, 0);
 
-    rb_eScriptError = rb_define_class("ScriptError", rb_eException);
-    rb_eSyntaxError = rb_define_class("SyntaxError", rb_eScriptError);
-    rb_eLoadError   = rb_define_class("LoadError", rb_eScriptError);
-    rb_eNotImpError = rb_define_class("NotImplementedError", rb_eScriptError);
-
     rb_eRuntimeError = rb_define_class("RuntimeError", rb_eStandardError);
-    rb_eSecurityError = rb_define_class("SecurityError", rb_eStandardError);
+    rb_eSecurityError = rb_define_class("SecurityError", rb_eException);
     rb_eNoMemError = rb_define_class("NoMemoryError", rb_eException);
 
     syserr_tbl = st_init_numtable();
@@ -1035,78 +1063,58 @@ Init_Exception()
     rb_define_singleton_method(rb_eSystemCallError, "===", syserr_eqq, 1);
 
     rb_mErrno = rb_define_module("Errno");
+    rb_define_singleton_method(rb_mErrno, "const_missing", errno_missing, 1);
 
     rb_define_global_function("warn", rb_warn_m, 1);
 }
 
 void
-#ifdef HAVE_STDARG_PROTOTYPES
 rb_raise(VALUE exc, const char *fmt, ...)
-#else
-rb_raise(exc, fmt, va_alist)
-    VALUE exc;
-    const char *fmt;
-    va_dcl
-#endif
 {
     va_list args;
-    char buf[BUFSIZ];
+    VALUE mesg;
 
-    va_init_list(args,fmt);
-    vsnprintf(buf, BUFSIZ, fmt, args);
+    va_start(args, fmt);
+    mesg = rb_vsprintf(fmt, args);
     va_end(args);
-    rb_exc_raise(rb_exc_new2(exc, buf));
+    rb_exc_raise(rb_exc_new3(exc, mesg));
 }
 
 void
-#ifdef HAVE_STDARG_PROTOTYPES
 rb_loaderror(const char *fmt, ...)
-#else
-rb_loaderror(fmt, va_alist)
-    const char *fmt;
-    va_dcl
-#endif
 {
     va_list args;
-    char buf[BUFSIZ];
+    VALUE mesg;
 
-    va_init_list(args, fmt);
-    vsnprintf(buf, BUFSIZ, fmt, args);
+    va_start(args, fmt);
+    mesg = rb_vsprintf(fmt, args);
     va_end(args);
-    rb_exc_raise(rb_exc_new2(rb_eLoadError, buf));
+    rb_exc_raise(rb_exc_new3(rb_eLoadError, mesg));
 }
 
 void
-rb_notimplement()
+rb_notimplement(void)
 {
     rb_raise(rb_eNotImpError,
 	     "%s() function is unimplemented on this machine",
-	     rb_id2name(ruby_frame->last_func));
+	     rb_id2name(rb_frame_this_func()));
 }
 
 void
-#ifdef HAVE_STDARG_PROTOTYPES
 rb_fatal(const char *fmt, ...)
-#else
-rb_fatal(fmt, va_alist)
-    const char *fmt;
-    va_dcl
-#endif
 {
     va_list args;
-    char buf[BUFSIZ];
+    VALUE mesg;
 
-    va_init_list(args, fmt);
-    vsnprintf(buf, BUFSIZ, fmt, args);
+    va_start(args, fmt);
+    mesg = rb_vsprintf(fmt, args);
     va_end(args);
 
-    ruby_in_eval = 0;
-    rb_exc_fatal(rb_exc_new2(rb_eFatal, buf));
+    rb_exc_fatal(rb_exc_new3(rb_eFatal, mesg));
 }
 
 void
-rb_sys_fail(mesg)
-    const char *mesg;
+rb_sys_fail(const char *mesg)
 {
     int n = errno;
     VALUE arg;
@@ -1121,54 +1129,45 @@ rb_sys_fail(mesg)
 }
 
 void
-#ifdef HAVE_STDARG_PROTOTYPES
 rb_sys_warning(const char *fmt, ...)
-#else
-rb_sys_warning(fmt, va_alist)
-     const char *fmt;
-     va_dcl
-#endif
 {
-     char buf[BUFSIZ];
-     va_list args;
-     int errno_save;
-     
-     errno_save = errno;
+    char buf[BUFSIZ];
+    va_list args;
+    int errno_save;
 
-     if (!RTEST(ruby_verbose)) return;
+    errno_save = errno;
 
-     snprintf(buf, BUFSIZ, "warning: %s", fmt);
-     snprintf(buf+strlen(buf), BUFSIZ-strlen(buf), ": %s", strerror(errno_save));
-     
-     va_init_list(args, fmt);
-     warn_print(buf, args);
-     va_end(args);
-     errno = errno_save;
+    if (!RTEST(ruby_verbose)) return;
+
+    snprintf(buf, BUFSIZ, "warning: %s", fmt);
+    snprintf(buf+strlen(buf), BUFSIZ-strlen(buf), ": %s", strerror(errno_save));
+
+    va_start(args, fmt);
+    warn_print(buf, args);
+    va_end(args);
+    errno = errno_save;
 }
 
 void
-rb_load_fail(path)
-    const char *path;
+rb_load_fail(const char *path)
 {
     rb_loaderror("%s -- %s", strerror(errno), path);
 }
 
 void
-rb_error_frozen(what)
-    const char *what;
+rb_error_frozen(const char *what)
 {
-    rb_raise(rb_eTypeError, "can't modify frozen %s", what);
+    rb_raise(rb_eRuntimeError, "can't modify frozen %s", what);
 }
 
 void
-rb_check_frozen(obj)
-    VALUE obj;
+rb_check_frozen(VALUE obj)
 {
     if (OBJ_FROZEN(obj)) rb_error_frozen(rb_obj_classname(obj));
 }
 
 void
-Init_syserr()
+Init_syserr(void)
 {
 #ifdef EPERM
     set_syserr(EPERM, "EPERM");
@@ -1536,27 +1535,33 @@ Init_syserr()
 #ifdef EDQUOT
     set_syserr(EDQUOT, "EDQUOT");
 #endif
+    rb_eNOERROR = set_syserr(0, "NOERROR");
 }
 
 static void
-err_append(s)
-    const char *s;
+err_append(const char *s)
 {
-    extern VALUE ruby_errinfo;
+    rb_thread_t *th = GET_THREAD();
+    VALUE err = th->errinfo;
 
-    if (ruby_in_eval) {
-	if (NIL_P(ruby_errinfo)) {
-	    ruby_errinfo = rb_exc_new2(rb_eSyntaxError, s);
+    if (th->mild_compile_error) {
+	if (!RTEST(err)) {
+	    err = rb_exc_new2(rb_eSyntaxError, s);
+	    th->errinfo = err;
 	}
 	else {
-	    VALUE str = rb_obj_as_string(ruby_errinfo);
+	    VALUE str = rb_obj_as_string(err);
 
 	    rb_str_cat2(str, "\n");
 	    rb_str_cat2(str, s);
-	    ruby_errinfo = rb_exc_new3(rb_eSyntaxError, str);
+	    th->errinfo = rb_exc_new3(rb_eSyntaxError, str);
 	}
     }
     else {
+	if (!RTEST(err)) {
+	    err = rb_exc_new2(rb_eSyntaxError, "compile error");
+	    th->errinfo = err;
+	}
 	rb_write_error(s);
 	rb_write_error("\n");
     }
