@@ -2,13 +2,13 @@
 /*
  * rubyext.c
  *
- * $Author: shyouhei $
- * $Date: 2007-02-13 08:01:19 +0900 (Tue, 13 Feb 2007) $
+ * $Author: nobu $
  *
  * Copyright (C) 2003-2005 why the lucky stiff
  */
 
-#include "ruby.h"
+#include "ruby/ruby.h"
+#include "ruby/encoding.h"
 #include "syck.h"
 #include <sys/types.h>
 #include <time.h>
@@ -49,8 +49,8 @@ typedef struct {
 /*
  * symbols and constants
  */
-static ID s_new, s_utc, s_at, s_to_f, s_to_i, s_read, s_binmode, s_call, s_cmp, s_transfer, s_update, s_dup, s_haskey, s_match, s_keys, s_unpack, s_tr_bang, s_default_set, s_tag_read_class, s_tag_subclasses, s_resolver, s_push, s_emitter, s_level, s_detect_implicit, s_node_import, s_out, s_input, s_intern, s_transform, s_yaml_new, s_yaml_initialize, s_node_export, s_to_yaml, s_write, s_set_resolver;
-static ID s_tags, s_domain, s_kind, s_name, s_options, s_type_id, s_type_id_set, s_style, s_style_set, s_value, s_value_set;
+static ID s_new, s_utc, s_at, s_to_f, s_to_i, s_read, s_binmode, s_call, s_cmp, s_transfer, s_update, s_dup, s_haskey, s_match, s_keys, s_unpack, s_tr_bang, s_default_set, s_tag_read_class, s_tag_subclasses, s_resolver, s_push, s_emitter, s_level, s_detect_implicit, s_node_import, s_out, s_input, s_intern, s_transform, s_yaml_new, s_yaml_initialize, s_node_export, s_to_yaml, s_write, s_set_resolver, s_each;
+static ID s_tags, s_kind, s_name, s_options, s_type_id, s_type_id_set, s_style, s_style_set, s_value, s_value_set;
 static VALUE sym_model, sym_generic, sym_input, sym_bytecode;
 static VALUE sym_scalar, sym_seq, sym_map;
 static VALUE sym_1quote, sym_2quote, sym_fold, sym_literal, sym_plain, sym_inline;
@@ -60,10 +60,10 @@ static VALUE oDefaultResolver, oGenericResolver;
 /*
  * my private collection of numerical oddities.
  */
-static double S_zero()    { return 0.0; }
-static double S_one() { return 1.0; }
-static double S_inf() { return S_one() / S_zero(); }
-static double S_nan() { return S_zero() / S_zero(); }
+static double S_zero(void)    { return 0.0; }
+static double S_one(void) { return 1.0; }
+static double S_inf(void) { return S_one() / S_zero(); }
+static double S_nan(void) { return S_zero() / S_zero(); }
 
 static VALUE syck_node_transform( VALUE );
 
@@ -71,7 +71,7 @@ static VALUE syck_node_transform( VALUE );
  * handler prototypes
  */
 SYMID rb_syck_load_handler _((SyckParser *, SyckNode *));
-void rb_syck_err_handler _((SyckParser *, char *));
+void rb_syck_err_handler _((SyckParser *, const char *));
 SyckNode * rb_syck_bad_anchor_handler _((SyckParser *, char *));
 void rb_syck_output_handler _((SyckEmitter *, char *, long));
 void rb_syck_emitter_handler _((SyckEmitter *, st_data_t));
@@ -97,14 +97,14 @@ struct emitter_xtra {
  * Convert YAML to bytecode
  */
 VALUE
-rb_syck_compile(self, port)
-    VALUE self, port;
+rb_syck_compile(VALUE self, VALUE port)
 {
     SYMID oid;
     int taint;
     char *ret;
     VALUE bc;
-    bytestring_t *sav; 
+    bytestring_t *sav = NULL;
+    void *data;
 
     SyckParser *parser = syck_new_parser();
     taint = syck_parser_assign_io(parser, &port);
@@ -113,7 +113,7 @@ rb_syck_compile(self, port)
     syck_parser_implicit_typing( parser, 0 );
     syck_parser_taguri_expansion( parser, 0 );
     oid = syck_parse( parser );
-    syck_lookup_sym( parser, oid, (char **)&sav );
+    if (syck_lookup_sym( parser, oid, &data )) sav = data;
 
     ret = S_ALLOCA_N( char, strlen( sav->buffer ) + 3 );
     ret[0] = '\0';
@@ -150,8 +150,8 @@ rb_syck_io_str_read( char *buf, SyckIoStr *str, long max_size, long skip )
         if (!NIL_P(str2))
         {
             StringValue(str2);
-            len = RSTRING(str2)->len;
-            memcpy( buf + skip, RSTRING(str2)->ptr, len );
+            len = RSTRING_LEN(str2);
+            memcpy( buf + skip, RSTRING_PTR(str2), len );
         }
     }
     len += skip;
@@ -164,16 +164,14 @@ rb_syck_io_str_read( char *buf, SyckIoStr *str, long max_size, long skip )
  * (returns tainted? boolean)
  */
 int
-syck_parser_assign_io(parser, pport)
-    SyckParser *parser;
-    VALUE *pport;
+syck_parser_assign_io(SyckParser *parser, VALUE *pport)
 {
     int taint = Qtrue;
     VALUE tmp, port = *pport;
     if (!NIL_P(tmp = rb_check_string_type(port))) {
         taint = OBJ_TAINTED(port); /* original taintedness */
         port = tmp;
-        syck_parser_str( parser, RSTRING(port)->ptr, RSTRING(port)->len, NULL );
+        syck_parser_str( parser, RSTRING_PTR(port), RSTRING_LEN(port), NULL );
     }
     else if (rb_respond_to(port, s_read)) {
         if (rb_respond_to(port, s_binmode)) {
@@ -192,8 +190,7 @@ syck_parser_assign_io(parser, pport)
  * Get value in hash by key, forcing an empty hash if nil.
  */
 VALUE
-syck_get_hash_aref(hsh, key)
-    VALUE hsh, key;
+syck_get_hash_aref(VALUE hsh, VALUE key)
 {
    VALUE val = rb_hash_aref( hsh, key );
    if ( NIL_P( val ) ) 
@@ -208,9 +205,7 @@ syck_get_hash_aref(hsh, key)
  * creating timestamps
  */
 SYMID
-rb_syck_mktime(str, len)
-    char *str;
-    long len;
+rb_syck_mktime(char *str, long len)
 {
     VALUE time;
     char *ptr = str;
@@ -268,9 +263,13 @@ rb_syck_mktime(str, len)
     {
         char padded[] = "000000";
         char *end = ptr + 1;
+        char *p = end;
         while ( isdigit( *end ) ) end++;
-        MEMCPY(padded, ptr + 1, char, end - (ptr + 1));
-        usec = strtol(padded, NULL, 10);
+        if (end - p < sizeof(padded)) {
+            MEMCPY(padded, ptr + 1, char, end - (ptr + 1));
+            p = padded;
+        }
+        usec = strtol(p, NULL, 10);
     }
     else
     {
@@ -315,8 +314,7 @@ rb_syck_mktime(str, len)
  * (see http://www.yaml.org/type/merge/)
  */
 VALUE
-syck_merge_i( entry, hsh )
-    VALUE entry, hsh;
+syck_merge_i(VALUE entry, VALUE hsh )
 {
     VALUE tmp;
     if ( !NIL_P(tmp = rb_check_convert_type(entry, T_HASH, "Hash", "to_hash")) )
@@ -331,9 +329,7 @@ syck_merge_i( entry, hsh )
  * default handler for ruby.yaml.org types
  */
 int
-yaml_org_handler( n, ref )
-    SyckNode *n;
-    VALUE *ref;
+yaml_org_handler( SyckNode *n, VALUE *ref )
 {
     char *type_id = n->type_id;
     int transferred = 0;
@@ -517,6 +513,7 @@ yaml_org_handler( n, ref )
             else if ( strcmp( type_id, "str" ) == 0 )
             {
                 obj = rb_str_new( n->data.str->ptr, n->data.str->len );
+                rb_enc_associate(obj, rb_utf8_encoding());
             }
             else
             {
@@ -571,7 +568,7 @@ yaml_org_handler( n, ref )
                             VALUE dup = rb_funcall( tmph, s_dup, 0 );
                             tmp = rb_ary_reverse( tmp );
                             rb_ary_push( tmp, obj );
-                            rb_iterate( rb_each, tmp, syck_merge_i, dup );
+                            rb_block_call( tmp, s_each, 0, 0, syck_merge_i, dup );
                             obj = dup;
                             skip_aset = 1;
                         }
@@ -602,9 +599,7 @@ static void syck_node_mark( SyckNode *n );
  * - Converts data into native Ruby types
  */
 SYMID
-rb_syck_load_handler(p, n)
-    SyckParser *p;
-    SyckNode *n;
+rb_syck_load_handler(SyckParser *p, SyckNode *n)
 {
     VALUE obj = Qnil;
     struct parser_xtra *bonus = (struct parser_xtra *)p->bonus;
@@ -632,7 +627,7 @@ rb_syck_load_handler(p, n)
     if ( bonus->taint)      OBJ_TAINT( obj );
     if ( bonus->proc != 0 ) rb_funcall(bonus->proc, s_call, 1, obj);
 
-    rb_hash_aset(bonus->data, INT2FIX(RHASH(bonus->data)->tbl->num_entries), obj);
+    rb_hash_aset(bonus->data, INT2FIX(RHASH_SIZE(bonus->data)), obj);
     return obj;
 }
 
@@ -640,9 +635,7 @@ rb_syck_load_handler(p, n)
  * friendly errors.
  */
 void
-rb_syck_err_handler(p, msg)
-    SyckParser *p;
-    char *msg;
+rb_syck_err_handler(SyckParser *p, const char *msg)
 {
     char *endl = p->cursor;
 
@@ -661,9 +654,7 @@ rb_syck_err_handler(p, msg)
  * provide bad anchor object to the parser.
  */
 SyckNode *
-rb_syck_bad_anchor_handler(p, a)
-    SyckParser *p;
-    char *a;
+rb_syck_bad_anchor_handler(SyckParser *p, char *a)
 {
     VALUE anchor_name = rb_str_new2( a );
     SyckNode *badanc = syck_new_map( rb_str_new2( "name" ), anchor_name );
@@ -675,8 +666,7 @@ rb_syck_bad_anchor_handler(p, a)
  * data loaded based on the model requested.
  */
 void
-syck_set_model( p, input, model )
-    VALUE p, input, model;
+syck_set_model(VALUE p, VALUE input, VALUE model)
 {
     SyckParser *parser;
     Data_Get_Struct(p, SyckParser, parser);
@@ -716,8 +706,7 @@ syck_st_mark_nodes( char *key, SyckNode *n, char *arg )
  * mark parser nodes
  */
 static void
-syck_mark_parser(parser)
-    SyckParser *parser;
+syck_mark_parser(SyckParser *parser)
 {
     struct parser_xtra *bonus = (struct parser_xtra *)parser->bonus;
     rb_gc_mark_maybe(parser->root);
@@ -740,8 +729,7 @@ syck_mark_parser(parser)
  * Free the parser and any bonus attachment.
  */
 void
-rb_syck_free_parser(p)
-    SyckParser *p;
+rb_syck_free_parser(SyckParser *p)
 {
     S_FREE( p->bonus );
     syck_free_parser(p);
@@ -752,8 +740,7 @@ rb_syck_free_parser(p)
  */
 VALUE syck_parser_s_alloc _((VALUE));
 VALUE 
-syck_parser_s_alloc(class)
-    VALUE class;
+syck_parser_s_alloc(VALUE class)
 {
     VALUE pobj;
     SyckParser *parser = syck_new_parser();
@@ -772,10 +759,7 @@ syck_parser_s_alloc(class)
  * YAML::Syck::Parser.initialize( resolver, options )
  */
 static VALUE
-syck_parser_initialize(argc, argv, self)
-    int argc;
-    VALUE *argv;
-    VALUE self;
+syck_parser_initialize(int argc, VALUE *argv, VALUE self)
 {
     VALUE options;
     if (rb_scan_args(argc, argv, "01", &options) == 0)
@@ -795,8 +779,7 @@ syck_parser_initialize(argc, argv, self)
  * YAML::Syck::Parser.bufsize = Integer
  */
 static VALUE
-syck_parser_bufsize_set( self, size )
-    VALUE self, size;
+syck_parser_bufsize_set(VALUE self, VALUE size)
 {
     SyckParser *parser;
 
@@ -812,8 +795,7 @@ syck_parser_bufsize_set( self, size )
  * YAML::Syck::Parser.bufsize => Integer
  */
 static VALUE
-syck_parser_bufsize_get( self )
-    VALUE self;
+syck_parser_bufsize_get(VALUE self)
 {
     SyckParser *parser;
 
@@ -825,10 +807,7 @@ syck_parser_bufsize_get( self )
  * YAML::Syck::Parser.load( IO or String )
  */
 VALUE
-syck_parser_load(argc, argv, self)
-    int argc;
-    VALUE *argv;
-    VALUE self;
+syck_parser_load(int argc, VALUE *argv, VALUE self)
 {
     VALUE port, proc, model, input;
     SyckParser *parser;
@@ -855,10 +834,7 @@ syck_parser_load(argc, argv, self)
  * YAML::Syck::Parser.load_documents( IO or String ) { |doc| }
  */
 VALUE
-syck_parser_load_documents(argc, argv, self)
-    int argc;
-    VALUE *argv;
-    VALUE self;
+syck_parser_load_documents(int argc, VALUE *argv, VALUE self)
 {
     VALUE port, proc, v, input, model;
     SyckParser *parser;
@@ -899,8 +875,7 @@ syck_parser_load_documents(argc, argv, self)
  * YAML::Syck::Parser#set_resolver
  */
 VALUE
-syck_parser_set_resolver( self, resolver )
-    VALUE self, resolver;
+syck_parser_set_resolver(VALUE self, VALUE resolver)
 {
     rb_ivar_set( self, s_resolver, resolver );
     return self;
@@ -910,10 +885,8 @@ syck_parser_set_resolver( self, resolver )
  * YAML::Syck::Resolver.initialize
  */
 static VALUE
-syck_resolver_initialize( self )
-    VALUE self;
+syck_resolver_initialize(VALUE self)
 {
-    VALUE tags = rb_hash_new();
     rb_ivar_set(self, s_tags, rb_hash_new());
     return self;
 }
@@ -922,8 +895,7 @@ syck_resolver_initialize( self )
  * YAML::Syck::Resolver#add_type
  */
 VALUE
-syck_resolver_add_type( self, taguri, cls )
-    VALUE self, taguri, cls;
+syck_resolver_add_type(VALUE self, VALUE taguri, VALUE cls)
 {
     VALUE tags = rb_attr_get(self, s_tags);
     rb_hash_aset( tags, taguri, cls );
@@ -934,8 +906,7 @@ syck_resolver_add_type( self, taguri, cls )
  * YAML::Syck::Resolver#use_types_at
  */
 VALUE
-syck_resolver_use_types_at( self, hsh )
-    VALUE self, hsh;
+syck_resolver_use_types_at(VALUE self, VALUE hsh)
 {
     rb_ivar_set( self, s_tags, hsh );
     return Qnil;
@@ -945,10 +916,8 @@ syck_resolver_use_types_at( self, hsh )
  * YAML::Syck::Resolver#detect_implicit 
  */
 VALUE
-syck_resolver_detect_implicit( self, val )
-    VALUE self, val;
+syck_resolver_detect_implicit(VALUE self, VALUE val)
 {
-    char *type_id;
     return rb_str_new2( "" );
 }
 
@@ -956,11 +925,10 @@ syck_resolver_detect_implicit( self, val )
  * YAML::Syck::Resolver#node_import
  */
 VALUE
-syck_resolver_node_import( self, node )
-    VALUE self, node;
+syck_resolver_node_import(VALUE self, VALUE node)
 {
     SyckNode *n;
-    VALUE obj;
+    VALUE obj = Qnil;
     int i = 0;
     Data_Get_Struct(node, SyckNode, n);
 
@@ -1006,7 +974,7 @@ syck_resolver_node_import( self, node )
                             VALUE dup = rb_funcall( end, s_dup, 0 );
                             v = rb_ary_reverse( v );
                             rb_ary_push( v, obj );
-                            rb_iterate( rb_each, v, syck_merge_i, dup );
+                            rb_block_call( v, s_each, 0, 0, syck_merge_i, dup );
                             obj = dup;
                             skip_aset = 1;
                         }
@@ -1037,16 +1005,15 @@ syck_resolver_node_import( self, node )
  * Set instance variables
  */
 VALUE
-syck_set_ivars( vars, obj )
-        VALUE vars, obj;
+syck_set_ivars(VALUE vars, VALUE obj)
 {
     VALUE ivname = rb_ary_entry( vars, 0 );
     char *ivn;
     StringValue( ivname );
-    ivn = S_ALLOCA_N( char, RSTRING(ivname)->len + 2 );
+    ivn = S_ALLOCA_N( char, RSTRING_LEN(ivname) + 2 );
     ivn[0] = '@';
     ivn[1] = '\0';
-    strncat( ivn, RSTRING(ivname)->ptr, RSTRING(ivname)->len );
+    strncat( ivn, RSTRING_PTR(ivname), RSTRING_LEN(ivname) );
     rb_iv_set( obj, ivn, rb_ary_entry( vars, 1 ) );
     return Qnil;
 }
@@ -1055,13 +1022,12 @@ syck_set_ivars( vars, obj )
  * YAML::Syck::Resolver#const_find
  */
 VALUE
-syck_const_find( const_name )
-    VALUE const_name;
+syck_const_find(VALUE const_name)
 {
     VALUE tclass = rb_cObject;
     VALUE tparts = rb_str_split( const_name, "::" );
     int i = 0;
-    for ( i = 0; i < RARRAY(tparts)->len; i++ ) {
+    for ( i = 0; i < RARRAY_LEN(tparts); i++ ) {
         VALUE tpart = rb_to_id( rb_ary_entry( tparts, i ) );
         if ( !rb_const_defined( tclass, tpart ) ) return Qnil;
         tclass = rb_const_get( tclass, tpart );
@@ -1073,15 +1039,14 @@ syck_const_find( const_name )
  * YAML::Syck::Resolver#transfer
  */
 VALUE
-syck_resolver_transfer( self, type, val )
-    VALUE self, type, val;
+syck_resolver_transfer(VALUE self, VALUE type, VALUE val)
 {
-    if (NIL_P(type) || RSTRING(StringValue(type))->len == 0) 
+    if (NIL_P(type) || RSTRING_LEN(StringValue(type)) == 0) 
     {
         type = rb_funcall( self, s_detect_implicit, 1, val );
     }
 
-    if ( ! (NIL_P(type) || RSTRING(StringValue(type))->len == 0) )
+    if ( ! (NIL_P(type) || RSTRING_LEN(StringValue(type)) == 0) )
     {
         VALUE str_xprivate = rb_str_new2( "x-private" );
         VALUE colon = rb_str_new2( ":" );
@@ -1098,7 +1063,7 @@ syck_resolver_transfer( self, type, val )
             VALUE subclass_parts = rb_ary_new();
             VALUE parts = rb_str_split( type, ":" );
 
-            while ( RARRAY(parts)->len > 1 )
+            while ( RARRAY_LEN(parts) > 1 )
             {
                 VALUE partial;
                 rb_ary_unshift( subclass_parts, rb_ary_pop( parts ) );
@@ -1116,7 +1081,7 @@ syck_resolver_transfer( self, type, val )
                 if ( ! NIL_P( target_class ) )
                 {
                     subclass = target_class;
-                    if ( RARRAY(subclass_parts)->len > 0 && rb_respond_to( target_class, s_tag_subclasses ) &&
+                    if ( RARRAY_LEN(subclass_parts) > 0 && rb_respond_to( target_class, s_tag_subclasses ) &&
                          RTEST( rb_funcall( target_class, s_tag_subclasses, 0 ) ) )
                     {
                         VALUE subclass_v;
@@ -1175,7 +1140,7 @@ syck_resolver_transfer( self, type, val )
                 }
                 else if ( !NIL_P( obj ) && rb_obj_is_instance_of( val, rb_cHash ) )
                 {
-                    rb_iterate( rb_each, val, syck_set_ivars, obj );
+                    rb_block_call( val, s_each, 0, 0, syck_set_ivars, obj );
                 }
             }
             else 
@@ -1205,14 +1170,13 @@ syck_resolver_transfer( self, type, val )
  * YAML::Syck::Resolver#tagurize
  */
 VALUE
-syck_resolver_tagurize( self, val )
-    VALUE self, val;
+syck_resolver_tagurize(VALUE self, VALUE val)
 {
     VALUE tmp = rb_check_string_type(val);
 
     if ( !NIL_P(tmp) )
     {
-        char *taguri = syck_type_id_to_uri( RSTRING(tmp)->ptr );
+        char *taguri = syck_type_id_to_uri( RSTRING_PTR(tmp) );
         val = rb_str_new2( taguri );
         S_FREE( taguri );
     }
@@ -1224,16 +1188,15 @@ syck_resolver_tagurize( self, val )
  * YAML::Syck::DefaultResolver#detect_implicit 
  */
 VALUE
-syck_defaultresolver_detect_implicit( self, val )
-    VALUE self, val;
+syck_defaultresolver_detect_implicit(VALUE self, VALUE val)
 {
-    char *type_id;
+    const char *type_id;
     VALUE tmp = rb_check_string_type(val);
 
     if ( !NIL_P(tmp) )
     {
         val = tmp;
-        type_id = syck_match_implicit( RSTRING(val)->ptr, RSTRING(val)->len );
+        type_id = syck_match_implicit( RSTRING_PTR(val), RSTRING_LEN(val) );
         return rb_str_new2( type_id );
     }
 
@@ -1244,8 +1207,7 @@ syck_defaultresolver_detect_implicit( self, val )
  * YAML::Syck::DefaultResolver#node_import
  */
 VALUE
-syck_defaultresolver_node_import( self, node )
-    VALUE self, node;
+syck_defaultresolver_node_import(VALUE self, VALUE node)
 {
     SyckNode *n;
     VALUE obj;
@@ -1261,8 +1223,7 @@ syck_defaultresolver_node_import( self, node )
  * YAML::Syck::GenericResolver#node_import
  */
 VALUE
-syck_genericresolver_node_import( self, node )
-    VALUE self, node;
+syck_genericresolver_node_import(VALUE self, VALUE node)
 {
     SyckNode *n;
     int i = 0;
@@ -1279,6 +1240,7 @@ syck_genericresolver_node_import( self, node )
         case syck_str_kind:
         {
             v = rb_str_new( n->data.str->ptr, n->data.str->len );
+	    rb_enc_associate(v, rb_utf8_encoding());
             if ( n->data.str->style == scalar_1quote )
             {
                 style = sym_1quote;
@@ -1304,7 +1266,6 @@ syck_genericresolver_node_import( self, node )
         break;
 
         case syck_seq_kind:
-            rb_iv_set(obj, "@kind", sym_seq);
             v = rb_ary_new2( syck_seq_count( n ) );
             for ( i = 0; i < syck_seq_count( n ); i++ )
             {
@@ -1315,10 +1276,10 @@ syck_genericresolver_node_import( self, node )
                 style = sym_inline;
             } 
             obj = rb_funcall( cSeq, s_new, 3, t, v, style );
+            rb_iv_set(obj, "@kind", sym_seq);
         break;
 
         case syck_map_kind:
-            rb_iv_set(obj, "@kind", sym_map);
             v = rb_hash_new();
             for ( i = 0; i < syck_map_count( n ); i++ )
             {
@@ -1329,6 +1290,7 @@ syck_genericresolver_node_import( self, node )
                 style = sym_inline;
             } 
             obj = rb_funcall( cMap, s_new, 3, t, v, style );
+            rb_iv_set(obj, "@kind", sym_map);
         break;
     }
 
@@ -1339,8 +1301,7 @@ syck_genericresolver_node_import( self, node )
  * YAML::Syck::BadAlias.initialize
  */
 VALUE
-syck_badalias_initialize( self, val )
-    VALUE self, val;
+syck_badalias_initialize(VALUE self, VALUE val)
 {
     rb_iv_set( self, "@name", val );
     return self;
@@ -1350,8 +1311,7 @@ syck_badalias_initialize( self, val )
  * YAML::Syck::BadAlias.<=>
  */
 VALUE
-syck_badalias_cmp( alias1, alias2 )
-    VALUE alias1, alias2;
+syck_badalias_cmp(VALUE alias1, VALUE alias2)
 {
     VALUE str1 = rb_ivar_get( alias1, s_name ); 
     VALUE str2 = rb_ivar_get( alias2, s_name ); 
@@ -1363,8 +1323,7 @@ syck_badalias_cmp( alias1, alias2 )
  * YAML::DomainType.initialize
  */
 VALUE
-syck_domaintype_initialize( self, domain, type_id, val )
-    VALUE self, domain, type_id, val;
+syck_domaintype_initialize(VALUE self, VALUE domain, VALUE type_id, VALUE val)
 {
     rb_iv_set( self, "@domain", domain );
     rb_iv_set( self, "@type_id", type_id );
@@ -1376,8 +1335,7 @@ syck_domaintype_initialize( self, domain, type_id, val )
  * YAML::Object.initialize
  */
 VALUE
-syck_yobject_initialize( self, klass, ivars )
-    VALUE self, klass, ivars;
+syck_yobject_initialize(VALUE self, VALUE klass, VALUE ivars)
 {
     rb_iv_set( self, "@class", klass );
     rb_iv_set( self, "@ivars", ivars );
@@ -1388,8 +1346,7 @@ syck_yobject_initialize( self, klass, ivars )
  * YAML::PrivateType.initialize
  */
 VALUE
-syck_privatetype_initialize( self, type_id, val )
-    VALUE self, type_id, val;
+syck_privatetype_initialize(VALUE self, VALUE type_id, VALUE val)
 {
     rb_iv_set( self, "@type_id", type_id );
     rb_iv_set( self, "@value", val );
@@ -1400,8 +1357,7 @@ syck_privatetype_initialize( self, type_id, val )
  * Mark node contents.
  */
 static void
-syck_node_mark( n )
-    SyckNode *n;
+syck_node_mark(SyckNode *n)
 {
     int i;
     rb_gc_mark_maybe( n->id );
@@ -1421,6 +1377,11 @@ syck_node_mark( n )
                 rb_gc_mark( syck_map_read( n, map_value, i ) );
             }
         break;
+
+	case syck_str_kind:
+	default:
+	    /* nothing */
+	break;
     }
 #if 0 /* maybe needed */
     if ( n->shortcut ) syck_node_mark( n->shortcut ); /* caution: maybe cyclic */
@@ -1431,8 +1392,7 @@ syck_node_mark( n )
  * YAML::Syck::Scalar.allocate
  */
 VALUE
-syck_scalar_alloc( class )
-    VALUE class;
+syck_scalar_alloc(VALUE class)
 {
     SyckNode *node = syck_alloc_str();
     VALUE obj = Data_Wrap_Struct( class, syck_node_mark, syck_free_node, node );
@@ -1444,8 +1404,7 @@ syck_scalar_alloc( class )
  * YAML::Syck::Scalar.initialize
  */
 VALUE
-syck_scalar_initialize( self, type_id, val, style )
-    VALUE self, type_id, val, style;
+syck_scalar_initialize(VALUE self, VALUE type_id, VALUE val, VALUE style)
 {
     rb_iv_set( self, "@kind", sym_scalar );
     rb_funcall( self, s_type_id_set, 1, type_id );
@@ -1458,8 +1417,7 @@ syck_scalar_initialize( self, type_id, val, style )
  * YAML::Syck::Scalar.style=
  */
 VALUE
-syck_scalar_style_set( self, style )
-    VALUE self, style;
+syck_scalar_style_set(VALUE self, VALUE style)
 {
     SyckNode *node;
     Data_Get_Struct( self, SyckNode, node );
@@ -1497,15 +1455,14 @@ syck_scalar_style_set( self, style )
  * YAML::Syck::Scalar.value=
  */
 VALUE
-syck_scalar_value_set( self, val )
-    VALUE self, val;
+syck_scalar_value_set(VALUE  self, VALUE val)
 {
     SyckNode *node;
     Data_Get_Struct( self, SyckNode, node );
 
     StringValue( val );
-    node->data.str->ptr = syck_strndup( RSTRING(val)->ptr, RSTRING(val)->len );
-    node->data.str->len = RSTRING(val)->len;
+    node->data.str->ptr = syck_strndup( RSTRING_PTR(val), RSTRING_LEN(val) );
+    node->data.str->len = RSTRING_LEN(val);
     node->data.str->style = scalar_none;
 
     rb_iv_set( self, "@value", val );
@@ -1516,8 +1473,7 @@ syck_scalar_value_set( self, val )
  * YAML::Syck::Seq.allocate
  */
 VALUE
-syck_seq_alloc( class )
-    VALUE class;
+syck_seq_alloc(VALUE class)
 {
     SyckNode *node;
     VALUE obj;
@@ -1531,8 +1487,7 @@ syck_seq_alloc( class )
  * YAML::Syck::Seq.initialize
  */
 VALUE
-syck_seq_initialize( self, type_id, val, style )
-    VALUE self, type_id, val, style;
+syck_seq_initialize(VALUE self, VALUE type_id, VALUE val, VALUE style)
 {
     SyckNode *node;
     Data_Get_Struct( self, SyckNode, node );
@@ -1548,8 +1503,7 @@ syck_seq_initialize( self, type_id, val, style )
  * YAML::Syck::Seq.value=
  */
 VALUE
-syck_seq_value_set( self, val )
-    VALUE self, val;
+syck_seq_value_set(VALUE self, VALUE val)
 {
     SyckNode *node;
     Data_Get_Struct( self, SyckNode, node );
@@ -1558,7 +1512,7 @@ syck_seq_value_set( self, val )
     if ( !NIL_P( val ) ) {
         int i;
         syck_seq_empty( node );
-        for ( i = 0; i < RARRAY( val )->len; i++ )
+        for ( i = 0; i < RARRAY_LEN( val ); i++ )
         {
             syck_seq_add( node, rb_ary_entry(val, i) );
         }
@@ -1572,8 +1526,7 @@ syck_seq_value_set( self, val )
  * YAML::Syck::Seq.add
  */
 VALUE
-syck_seq_add_m( self, val )
-    VALUE self, val;
+syck_seq_add_m(VALUE self, VALUE val)
 {
     SyckNode *node;
     VALUE emitter = rb_ivar_get( self, s_emitter );
@@ -1592,8 +1545,7 @@ syck_seq_add_m( self, val )
  * YAML::Syck::Seq.style=
  */
 VALUE
-syck_seq_style_set( self, style )
-    VALUE self, style;
+syck_seq_style_set(VALUE self, VALUE style)
 {
     SyckNode *node;
     Data_Get_Struct( self, SyckNode, node );
@@ -1615,8 +1567,7 @@ syck_seq_style_set( self, style )
  * YAML::Syck::Map.allocate
  */
 VALUE
-syck_map_alloc( class )
-    VALUE class;
+syck_map_alloc(VALUE class)
 {
     SyckNode *node;
     VALUE obj;
@@ -1630,8 +1581,7 @@ syck_map_alloc( class )
  * YAML::Syck::Map.initialize
  */
 VALUE
-syck_map_initialize( self, type_id, val, style )
-    VALUE self, type_id, val, style;
+syck_map_initialize(VALUE self, VALUE type_id, VALUE val, VALUE style)
 {
     SyckNode *node;
     Data_Get_Struct( self, SyckNode, node );
@@ -1647,7 +1597,7 @@ syck_map_initialize( self, type_id, val, style )
         }
 
         keys = rb_funcall( hsh, s_keys, 0 );
-        for ( i = 0; i < RARRAY(keys)->len; i++ )
+        for ( i = 0; i < RARRAY_LEN(keys); i++ )
         {
             VALUE key = rb_ary_entry(keys, i);
             syck_map_add( node, key, rb_hash_aref(hsh, key) );
@@ -1665,8 +1615,7 @@ syck_map_initialize( self, type_id, val, style )
  * YAML::Syck::Map.value=
  */
 VALUE
-syck_map_value_set( self, val )
-    VALUE self, val;
+syck_map_value_set(VALUE self, VALUE val)
 {
     SyckNode *node;
     Data_Get_Struct( self, SyckNode, node );
@@ -1683,7 +1632,7 @@ syck_map_value_set( self, val )
 
         syck_map_empty( node );
         keys = rb_funcall( hsh, s_keys, 0 );
-        for ( i = 0; i < RARRAY(keys)->len; i++ )
+        for ( i = 0; i < RARRAY_LEN(keys); i++ )
         {
             VALUE key = rb_ary_entry(keys, i);
             syck_map_add( node, key, rb_hash_aref(hsh, key) );
@@ -1698,8 +1647,7 @@ syck_map_value_set( self, val )
  * YAML::Syck::Map.add
  */
 VALUE
-syck_map_add_m( self, key, val )
-    VALUE self, key, val;
+syck_map_add_m(VALUE self, VALUE key, VALUE val)
 {
     SyckNode *node;
     VALUE emitter = rb_ivar_get( self, s_emitter );
@@ -1719,8 +1667,7 @@ syck_map_add_m( self, key, val )
  * YAML::Syck::Map.style=
  */
 VALUE
-syck_map_style_set( self, style )
-    VALUE self, style;
+syck_map_style_set(VALUE self, VALUE style)
 {
     SyckNode *node;
     Data_Get_Struct( self, SyckNode, node );
@@ -1742,8 +1689,7 @@ syck_map_style_set( self, style )
  * Cloning method for all node types
  */
 VALUE
-syck_node_init_copy( copy, orig )
-    VALUE copy, orig;
+syck_node_init_copy(VALUE copy, VALUE orig)
 {
     SyckNode *copy_n;
     SyckNode *orig_n;
@@ -1766,8 +1712,7 @@ syck_node_init_copy( copy, orig )
  * YAML::Syck::Node#type_id=
  */
 VALUE
-syck_node_type_id_set( self, type_id )
-    VALUE self, type_id;
+syck_node_type_id_set(VALUE self, VALUE type_id)
 {
     SyckNode *node;
     Data_Get_Struct( self, SyckNode, node );
@@ -1776,7 +1721,7 @@ syck_node_type_id_set( self, type_id )
 
     if ( !NIL_P( type_id ) ) {
         StringValue( type_id );
-        node->type_id = syck_strndup( RSTRING(type_id)->ptr, RSTRING(type_id)->len );
+        node->type_id = syck_strndup( RSTRING_PTR(type_id), RSTRING_LEN(type_id) );
     }
 
     rb_iv_set( self, "@type_id", type_id );
@@ -1787,11 +1732,10 @@ syck_node_type_id_set( self, type_id )
  * YAML::Syck::Node.transform
  */
 VALUE
-syck_node_transform( self )
-    VALUE self;
+syck_node_transform(VALUE self)
 {
     VALUE t;
-    SyckNode *n;
+    SyckNode *n = NULL;
     SyckNode *orig_n;
     Data_Get_Struct(self, SyckNode, orig_n);
     t = Data_Wrap_Struct( cNode, syck_node_mark, syck_free_node, 0 );
@@ -1844,9 +1788,7 @@ syck_node_transform( self )
  * No one could possibly object.
  */
 void
-rb_syck_emitter_handler(e, data)
-    SyckEmitter *e;
-    st_data_t data;
+rb_syck_emitter_handler(SyckEmitter *e, st_data_t data)
 {
     SyckNode *n;
     Data_Get_Struct((VALUE)data, SyckNode, n);
@@ -1890,10 +1832,7 @@ rb_syck_emitter_handler(e, data)
  * Handle output from the emitter
  */
 void 
-rb_syck_output_handler( emitter, str, len )
-    SyckEmitter *emitter;
-    char *str;
-    long len;
+rb_syck_output_handler(SyckEmitter * emitter, char *str, long len)
 {
     struct emitter_xtra *bonus = (struct emitter_xtra *)emitter->bonus;
     VALUE dest = bonus->port;
@@ -1909,8 +1848,7 @@ rb_syck_output_handler( emitter, str, len )
  * symbol table.
  */
 void
-syck_out_mark( emitter, node )
-    VALUE emitter, node;
+syck_out_mark(VALUE emitter, VALUE node)
 {
     SyckEmitter *emitterPtr;
     struct emitter_xtra *bonus;
@@ -1927,8 +1865,7 @@ syck_out_mark( emitter, node )
  * Mark emitter values.
  */
 static void
-syck_mark_emitter(emitter)
-    SyckEmitter *emitter;
+syck_mark_emitter(SyckEmitter *emitter)
 {
     struct emitter_xtra *bonus = (struct emitter_xtra *)emitter->bonus;
     rb_gc_mark( bonus->oid  );
@@ -1940,8 +1877,7 @@ syck_mark_emitter(emitter)
  * Free the emitter and any bonus attachment.
  */
 void
-rb_syck_free_emitter(e)
-    SyckEmitter *e;
+rb_syck_free_emitter(SyckEmitter *e)
 {
     S_FREE( e->bonus );
     syck_free_emitter(e);
@@ -1952,8 +1888,7 @@ rb_syck_free_emitter(e)
  */
 VALUE syck_emitter_s_alloc _((VALUE));
 VALUE 
-syck_emitter_s_alloc(class)
-    VALUE class;
+syck_emitter_s_alloc(VALUE class)
 {
     VALUE pobj;
     SyckEmitter *emitter = syck_new_emitter();
@@ -1973,10 +1908,7 @@ syck_emitter_s_alloc(class)
  * YAML::Syck::Emitter.reset( options )
  */
 VALUE
-syck_emitter_reset( argc, argv, self )
-    int argc;
-    VALUE *argv;
-    VALUE self;
+syck_emitter_reset(int argc, VALUE *argv, VALUE self)
 {
     VALUE options, tmp;
     SyckEmitter *emitter;
@@ -2018,13 +1950,9 @@ syck_emitter_reset( argc, argv, self )
  * YAML::Syck::Emitter.emit( object_id ) { |out| ... }
  */
 VALUE
-syck_emitter_emit( argc, argv, self )
-    int argc;
-    VALUE *argv;
-    VALUE self;
+syck_emitter_emit(int argc, VALUE *argv, VALUE self)
 {
     VALUE oid, proc;
-    char *anchor_name;
     SyckEmitter *emitter;
     struct emitter_xtra *bonus;
     SYMID symple;
@@ -2062,8 +1990,7 @@ syck_emitter_emit( argc, argv, self )
  * YAML::Syck::Emitter#node_export
  */
 VALUE
-syck_emitter_node_export( self, node )
-    VALUE self, node;
+syck_emitter_node_export(VALUE self, VALUE node)
 {
     return rb_funcall( node, s_to_yaml, 1, self );
 }
@@ -2072,8 +1999,7 @@ syck_emitter_node_export( self, node )
  * YAML::Syck::Emitter#set_resolver
  */
 VALUE
-syck_emitter_set_resolver( self, resolver )
-    VALUE self, resolver;
+syck_emitter_set_resolver(VALUE self, VALUE resolver)
 {
     rb_ivar_set( self, s_resolver, resolver );
     return self;
@@ -2083,8 +2009,7 @@ syck_emitter_set_resolver( self, resolver )
  * YAML::Syck::Out::initialize
  */
 VALUE
-syck_out_initialize( self, emitter )
-    VALUE self, emitter;
+syck_out_initialize(VALUE self, VALUE emitter)
 {
     rb_ivar_set( self, s_emitter, emitter );
     return self;
@@ -2094,10 +2019,7 @@ syck_out_initialize( self, emitter )
  * YAML::Syck::Out::map
  */
 VALUE
-syck_out_map( argc, argv, self )
-    int argc;
-    VALUE *argv;
-    VALUE self;
+syck_out_map(int argc, VALUE *argv, VALUE self)
 {
     VALUE type_id, style, map;
     if (rb_scan_args(argc, argv, "11", &type_id, &style) == 1) {
@@ -2113,10 +2035,7 @@ syck_out_map( argc, argv, self )
  * YAML::Syck::Out::seq
  */
 VALUE
-syck_out_seq( argc, argv, self )
-    int argc;
-    VALUE *argv;
-    VALUE self;
+syck_out_seq(int argc, VALUE *argv, VALUE self)
 {
     VALUE type_id, style, seq;
     if (rb_scan_args(argc, argv, "11", &type_id, &style) == 1) {
@@ -2134,15 +2053,10 @@ syck_out_scalar( self, type_id, str, style )
     VALUE self, type_id, str, style;
  */
 VALUE
-syck_out_scalar( argc, argv, self )
-    int argc;
-    VALUE *argv;
-    VALUE self;
+syck_out_scalar(int argc, VALUE *argv, VALUE self)
 {
     VALUE type_id, str, style, scalar;
-    if (rb_scan_args(argc, argv, "21", &type_id, &str, &style) == 2) {
-        style = Qnil;
-    }
+    rb_scan_args(argc, argv, "21", &type_id, &str, &style);
     scalar = rb_funcall( cScalar, s_new, 3, type_id, str, style );
     syck_out_mark( rb_ivar_get( self, s_emitter ), scalar );
     return scalar;
@@ -2194,6 +2108,7 @@ Init_syck()
     s_transform = rb_intern( "transform" );
     s_yaml_new = rb_intern("yaml_new");
     s_yaml_initialize = rb_intern("yaml_initialize");
+    s_each = rb_intern("each");
 
     s_tags = rb_intern("@tags");
     s_name = rb_intern("@name");
@@ -2280,7 +2195,6 @@ Init_syck()
      */
     cScalar = rb_define_class_under( rb_syck, "Scalar", cNode );
     rb_define_alloc_func( cScalar, syck_scalar_alloc );
-    rb_define_attr( cNode, "value", 1, 0 );
     rb_define_method( cScalar, "initialize", syck_scalar_initialize, 3 );
     rb_define_method( cScalar, "value=", syck_scalar_value_set, 1 );
     rb_define_method( cScalar, "style=", syck_scalar_style_set, 1 );
