@@ -284,7 +284,7 @@ class CGI
   # Standard internet newline sequence
   EOL = CR + LF
 
-  REVISION = '$Id: cgi.rb 27932 2010-05-20 19:14:58Z wyhaines $' #:nodoc:
+  REVISION = '$Id: cgi.rb 17753 2008-06-30 16:07:19Z nobu $' #:nodoc:
 
   NEEDS_BINMODE = true if /WIN/ni.match(RUBY_PLATFORM) 
 
@@ -306,7 +306,7 @@ class CGI
     "METHOD_NOT_ALLOWED"  => "405 Method Not Allowed",
     "NOT_ACCEPTABLE"      => "406 Not Acceptable",
     "LENGTH_REQUIRED"     => "411 Length Required",
-    "PRECONDITION_FAILED" => "412 Precondition Failed",
+    "PRECONDITION_FAILED" => "412 Rrecondition Failed",
     "SERVER_ERROR"        => "500 Internal Server Error",
     "NOT_IMPLEMENTED"     => "501 Method Not Implemented",
     "BAD_GATEWAY"         => "502 Bad Gateway",
@@ -339,8 +339,8 @@ class CGI
   #   url_encoded_string = CGI::escape("'Stop!' said Fred")
   #      # => "%27Stop%21%27+said+Fred"
   def CGI::escape(string)
-    string.gsub(/([^ a-zA-Z0-9_.-]+)/n) do
-      '%' + $1.unpack('H2' * $1.size).join('%').upcase
+    string.gsub(/([^ a-zA-Z0-9_.-]+)/) do
+      '%' + $1.unpack('H2' * $1.bytesize).join('%').upcase
     end.tr(' ', '+')
   end
 
@@ -349,17 +349,24 @@ class CGI
   #   string = CGI::unescape("%27Stop%21%27+said+Fred")
   #      # => "'Stop!' said Fred"
   def CGI::unescape(string)
-    string.tr('+', ' ').gsub(/((?:%[0-9a-fA-F]{2})+)/n) do
-      [$1.delete('%')].pack('H*')
+    enc = string.encoding
+    string.tr('+', ' ').gsub(/((?:%[0-9a-fA-F]{2})+)/) do
+      [$1.delete('%')].pack('H*').force_encoding(enc)
     end
   end
 
+  TABLE_FOR_ESCAPE_HTML__ = {
+    '&' => '&amp;',
+    '"' => '&quot;',
+    '<' => '&lt;',
+    '>' => '&gt;',
+  }
 
   # Escape special characters in HTML, namely &\"<>
   #   CGI::escapeHTML('Usage: foo "bar" <baz>')
   #      # => "Usage: foo &quot;bar&quot; &lt;baz&gt;"
   def CGI::escapeHTML(string)
-    string.gsub(/&/n, '&amp;').gsub(/\"/n, '&quot;').gsub(/>/n, '&gt;').gsub(/</n, '&lt;')
+    string.gsub(/[&\"<>]/, TABLE_FOR_ESCAPE_HTML__)
   end
 
 
@@ -367,39 +374,37 @@ class CGI
   #   CGI::unescapeHTML("Usage: foo &quot;bar&quot; &lt;baz&gt;")
   #      # => "Usage: foo \"bar\" <baz>"
   def CGI::unescapeHTML(string)
-    string.gsub(/&(amp|quot|gt|lt|\#[0-9]+|\#x[0-9A-Fa-f]+);/n) do
+    enc = string.encoding
+    string.gsub(/&(amp|quot|gt|lt|\#[0-9]+|\#x[0-9A-Fa-f]+);/) do
       match = $1.dup
       case match
       when 'amp'                 then '&'
       when 'quot'                then '"'
       when 'gt'                  then '>'
       when 'lt'                  then '<'
-      when /\A#0*(\d+)\z/n       then
+      when /\A#0*(\d+)\z/        then
         if Integer($1) < 256
-          Integer($1).chr
+          Integer($1).chr.force_encoding(enc)
         else
-          if Integer($1) < 65536 and ($KCODE[0] == ?u or $KCODE[0] == ?U)
-            [Integer($1)].pack("U")
-          else
-            "&##{$1};"
-          end
+          "&##{$1};"
         end
-      when /\A#x([0-9a-f]+)\z/ni then
+      when /\A#x([0-9a-f]+)\z/i then
         if $1.hex < 256
-          $1.hex.chr
+          $1.hex.chr.force_encoding(enc)
         else
-          if $1.hex < 65536 and ($KCODE[0] == ?u or $KCODE[0] == ?U)
-            [$1.hex].pack("U")
-          else
-            "&#x#{$1};"
-          end
+          "&#x#{$1};"
         end
       else
         "&#{match};"
       end
     end
   end
-
+  def CGI::escape_html(str)
+    escapeHTML(str)
+  end
+  def CGI::unescape_html(str)
+    unescapeHTML(str)
+  end
 
   # Escape only the tags of certain HTML elements in +string+.
   #
@@ -445,7 +450,12 @@ class CGI
       string
     end
   end
-
+  def CGI::escape_element(str)
+    escapeElement(str)
+  end
+  def CGI::unescape_element(str)
+    unescapeElement(str)
+  end
 
   # Format a +Time+ object as a String using the format specified by RFC 1123.
   #
@@ -709,13 +719,13 @@ class CGI
       require "nkf"
       case options["charset"]
       when /iso-2022-jp/ni
-        content = NKF::nkf('-m0 -x -j', content)
+        content = NKF::nkf('-j -m0 -x', content)
         options["language"] = "ja" unless options.has_key?("language")
       when /euc-jp/ni
-        content = NKF::nkf('-m0 -x -e', content)
+        content = NKF::nkf('-e -m0 -x', content)
         options["language"] = "ja" unless options.has_key?("language")
       when /shift_jis/ni
-        content = NKF::nkf('-m0 -x -s', content)
+        content = NKF::nkf('-s -m0 -x', content)
         options["language"] = "ja" unless options.has_key?("language")
       end
     end
@@ -792,11 +802,16 @@ class CGI
     #
     # These keywords correspond to attributes of the cookie object.
     def initialize(name = "", *value)
-      options = if name.kind_of?(String)
-                  { "name" => name, "value" => value }
-                else
-                  name
-                end
+      if name.kind_of?(String)
+        @name = name
+        @value = value
+        %r|^(.*/)|.match(ENV["SCRIPT_NAME"])
+        @path = ($1 or "")
+        @secure = false
+        return super(@value)
+      end
+
+      options = name
       unless options.has_key?("name")
         raise ArgumentError, "`name' required"
       end
@@ -817,8 +832,8 @@ class CGI
       super(@value)
     end
 
-    attr_accessor("name", "path", "domain", "expires")
-    attr_reader("secure", "value")
+    attr_accessor("name", "value", "path", "domain", "expires")
+    attr_reader("secure")
 
     # Set whether the Cookie is a secure cookie or not.
     #
@@ -828,17 +843,16 @@ class CGI
       @secure
     end
 
-    # Set the value of the cookie.
-    def value=(val)
-      @value.replace(Array(val))
-    end
-
     # Convert the Cookie to its string representation.
     def to_s
       buf = ""
       buf += @name + '='
 
-      buf += @value.map { |v| CGI::escape(v.to_s) }.join("&")
+      if @value.kind_of?(String)
+        buf += CGI::escape(@value)
+      else
+        buf += @value.collect{|v| CGI::escape(v) }.join("&")
+      end
 
       if @domain
         buf += '; domain=' + @domain
@@ -881,7 +895,7 @@ class CGI
       if cookies.has_key?(name)
         values = cookies[name].value + values
       end
-      cookies[name] = Cookie::new({ "name" => name, "value" => values })
+      cookies[name] = Cookie::new(name, *values)
     end
 
     cookies
@@ -954,11 +968,11 @@ class CGI
     end
 
     # Get the cookies as a hash of cookie-name=>Cookie pairs.
-    attr_accessor("cookies")
+    attr_accessor :cookies
 
     # Get the parameters as a hash of name=>values pairs, where
     # values is an Array.
-    attr("params")
+    attr_reader :params
 
     # Set all the parameters.
     def params=(hash)
@@ -969,7 +983,7 @@ class CGI
     def read_multipart(boundary, content_length)
       params = Hash.new([])
       boundary = "--" + boundary
-      quoted_boundary = Regexp.quote(boundary, "n")
+      quoted_boundary = Regexp.quote(boundary)
       buf = ""
       bufsize = 10 * 1024
       boundary_end=""
@@ -987,22 +1001,9 @@ class CGI
 
       loop do
         head = nil
-        if 10240 < content_length
-          require "tempfile"
-          body = Tempfile.new("CGI")
-        else
-          begin
-            require "stringio"
-            body = StringIO.new
-          rescue LoadError
-            require "tempfile"
-            body = Tempfile.new("CGI")
-          end
-        end
-        body.binmode if defined? body.binmode
+        body = MorphingBody.new
 
         until head and /#{quoted_boundary}(?:#{EOL}|--)/n.match(buf)
-
           if (not head) and /#{EOL}#{EOL}/n.match(buf)
             buf = buf.sub(/\A((?:.|\n)*?#{EOL})#{EOL}/n) do
               head = $1.dup
@@ -1057,7 +1058,7 @@ class CGI
         end
 
         /Content-Disposition:.* name="?([^\";\s]*)"?/ni.match(head)
-        name = $1.dup
+        name = ($1 || "").dup
 
         if params.has_key?(name)
           params[name].push(body)
@@ -1098,6 +1099,59 @@ class CGI
     end
     private :read_from_cmdline
 
+    # A wrapper class to use a StringIO object as the body and switch
+    # to a TempFile when the passed threshold is passed.
+    class MorphingBody
+      begin
+        require "stringio"
+        @@small_buffer = lambda{StringIO.new}
+      rescue LoadError
+        require "tempfile"
+        @@small_buffer = lambda{
+          n = Tempfile.new("CGI")
+          n.binmode
+          n
+        }
+      end
+
+      def initialize(morph_threshold = 10240)
+        @threshold = morph_threshold
+        @body = @@small_buffer.call
+        @cur_size = 0
+        @morph_check = true
+      end
+
+      def print(data)
+        if @morph_check && (@cur_size + data.size > @threshold)
+          convert_body
+        end
+        @body.print data
+      end
+      def rewind
+        @body.rewind
+      end
+      def path
+        @body.path
+      end
+
+      # returns the true body object.
+      def extract
+        @body
+      end
+
+      private
+      def convert_body
+        new_body = TempFile.new("CGI")
+        new_body.binmode if defined? @body.binmode
+        new_body.binmode if defined? new_body.binmode
+
+        @body.rewind
+        new_body.print @body.read
+        @body = new_body
+        @morph_check = false
+      end
+    end
+
     # Initialize the data from the query.
     #
     # Handles multipart forms (in particular, forms that involve file uploads).
@@ -1135,29 +1189,6 @@ class CGI
       @multipart
     end
 
-    module Value    # :nodoc:
-      def set_params(params)
-        @params = params
-      end
-      def [](idx, *args)
-        if args.size == 0
-          warn "#{caller(1)[0]}:CAUTION! cgi['key'] == cgi.params['key'][0]; if want Array, use cgi.params['key']"
-          @params[idx]
-        else
-          super[idx,*args]
-        end
-      end
-      def first
-        warn "#{caller(1)[0]}:CAUTION! cgi['key'] == cgi.params['key'][0]; if want Array, use cgi.params['key']"
-        self
-      end
-      alias last first
-      def to_a
-        @params || [self]
-      end
-      alias to_ary to_a   	# to be rhs of multiple assignment
-    end
-
     # Get the value for the parameter with a given key.
     #
     # If the parameter has multiple values, only the first will be 
@@ -1176,8 +1207,6 @@ class CGI
         end
       else
         str = if value then value.dup else "" end
-        str.extend(Value)
-        str.set_params(params)
         str
       end
     end
@@ -1259,7 +1288,7 @@ class CGI
             else
               '="' + CGI::escapeHTML(value) + '"'
             end
-          }.to_s + ">"
+          }.join + ">"
       END
       s.sub!(/\Z/, " +") << append if append
       s
@@ -1482,7 +1511,7 @@ class CGI
             value[value.size - 1]
           end
         end
-      }.to_s
+      }.join
     end
 
 
@@ -1562,7 +1591,7 @@ class CGI
       if @output_hidden
         body += @output_hidden.collect{|k,v|
           "<INPUT TYPE=\"HIDDEN\" NAME=\"#{k}\" VALUE=\"#{v}\">"
-        }.to_s
+        }.join
       end
       super(attributes){body}
     end
@@ -1860,7 +1889,7 @@ class CGI
               }
             end
           end
-        }.to_s
+        }.join
       }
 
     end
@@ -1937,7 +1966,7 @@ class CGI
             value[value.size - 1]
           end
         end
-      }.to_s
+      }.join
     end
 
     # Generate a reset button Input element, as a String.
@@ -2267,15 +2296,15 @@ class CGI
       Apache.request.setup_cgi_env
     end
 
+      (class << self; self; end).class_eval do
+        const_set(:CGI_PARAMS,  [1])
+        const_set(:CGI_COOKIES, [2])
+      end
+
     extend QueryExtension
     @multipart = false
-    if defined?(CGI_PARAMS)
-      warn "do not use CGI_PARAMS and CGI_COOKIES"
-      @params = CGI_PARAMS.dup
-      @cookies = CGI_COOKIES.dup
-    else
-      initialize_query()  # set @params, @cookies
-    end
+
+    initialize_query()  # set @params, @cookies
     @output_cookies = nil
     @output_hidden = nil
 
