@@ -15,14 +15,27 @@
 
 VALUE rb_mMath;
 
-#define Need_Float(x) (x) = rb_Float(x)
+static VALUE
+to_flo(VALUE x)
+{
+    if (!rb_obj_is_kind_of(x, rb_cNumeric)) {
+	rb_raise(rb_eTypeError, "can't convert %s into Float",
+		 NIL_P(x) ? "nil" :
+		 x == Qtrue ? "true" :
+		 x == Qfalse ? "false" :
+		 rb_obj_classname(x));
+    }
+    return rb_convert_type(x, T_FLOAT, "Float", "to_f");
+}
+
+#define Need_Float(x) (x) = to_flo(x)
 #define Need_Float2(x,y) do {\
     Need_Float(x);\
     Need_Float(y);\
 } while (0)
 
 static void
-domain_check(double x, char *msg)
+domain_check(double x, const char *msg)
 {
     while(1) {
 	if (errno) {
@@ -50,7 +63,7 @@ domain_check(double x, char *msg)
  *     
  */
 
-static VALUE
+VALUE
 math_atan2(VALUE obj, VALUE y, VALUE x)
 {
     Need_Float2(y, x);
@@ -66,7 +79,7 @@ math_atan2(VALUE obj, VALUE y, VALUE x)
  *  -1..1.
  */
 
-static VALUE
+VALUE
 math_cos(VALUE obj, VALUE x)
 {
     Need_Float(x);
@@ -81,7 +94,7 @@ math_cos(VALUE obj, VALUE x)
  *  -1..1.
  */
 
-static VALUE
+VALUE
 math_sin(VALUE obj, VALUE x)
 {
     Need_Float(x);
@@ -172,7 +185,7 @@ cosh(double x)
  *  Computes the hyperbolic cosine of <i>x</i> (expressed in radians).
  */
 
-static VALUE
+VALUE
 math_cosh(VALUE obj, VALUE x)
 {
     Need_Float(x);
@@ -196,7 +209,7 @@ sinh(double x)
  *  radians).
  */
 
-static VALUE
+VALUE
 math_sinh(VALUE obj, VALUE x)
 {
     Need_Float(x);
@@ -285,7 +298,7 @@ math_atanh(VALUE obj, VALUE x)
  *  Returns e**x.
  */
 
-static VALUE
+VALUE
 math_exp(VALUE obj, VALUE x)
 {
     Need_Float(x);
@@ -311,7 +324,7 @@ math_exp(VALUE obj, VALUE x)
  *  of logarithm.
  */
 
-static VALUE
+VALUE
 math_log(int argc, VALUE *argv)
 {
     VALUE x, base;
@@ -386,9 +399,26 @@ math_log10(VALUE obj, VALUE x)
  *     Math.sqrt(numeric)    => float
  *  
  *  Returns the non-negative square root of <i>numeric</i>.
+ *
+ *    0.upto(10) {|x|            
+ *      p [x, Math.sqrt(x), Math.sqrt(x)**2]
+ *    }
+ *    #=>
+ *    [0, 0.0, 0.0]
+ *    [1, 1.0, 1.0]
+ *    [2, 1.4142135623731, 2.0]
+ *    [3, 1.73205080756888, 3.0]
+ *    [4, 2.0, 4.0]
+ *    [5, 2.23606797749979, 5.0]
+ *    [6, 2.44948974278318, 6.0]
+ *    [7, 2.64575131106459, 7.0]
+ *    [8, 2.82842712474619, 8.0]
+ *    [9, 3.0, 9.0]
+ *    [10, 3.16227766016838, 10.0]
+ *
  */
 
-static VALUE
+VALUE
 math_sqrt(VALUE obj, VALUE x)
 {
     double d;
@@ -405,6 +435,31 @@ math_sqrt(VALUE obj, VALUE x)
  *     Math.cbrt(numeric)    => float
  *  
  *  Returns the cube root of <i>numeric</i>.
+ *
+ *    -9.upto(9) {|x|
+ *      p [x, Math.cbrt(x), Math.cbrt(x)**3]
+ *    }
+ *    #=>
+ *    [-9, -2.0800838230519, -9.0]
+ *    [-8, -2.0, -8.0]
+ *    [-7, -1.91293118277239, -7.0]
+ *    [-6, -1.81712059283214, -6.0]
+ *    [-5, -1.7099759466767, -5.0]
+ *    [-4, -1.5874010519682, -4.0]
+ *    [-3, -1.44224957030741, -3.0]
+ *    [-2, -1.25992104989487, -2.0]
+ *    [-1, -1.0, -1.0]
+ *    [0, 0.0, 0.0]
+ *    [1, 1.0, 1.0]
+ *    [2, 1.25992104989487, 2.0]
+ *    [3, 1.44224957030741, 3.0]
+ *    [4, 1.5874010519682, 4.0]
+ *    [5, 1.7099759466767, 5.0]
+ *    [6, 1.81712059283214, 6.0]
+ *    [7, 1.91293118277239, 7.0]
+ *    [8, 2.0, 8.0]
+ *    [9, 2.0800838230519, 9.0]
+ *
  */
 
 static VALUE
@@ -465,7 +520,7 @@ math_ldexp(VALUE obj, VALUE x, VALUE n)
  *     Math.hypot(3, 4)   #=> 5.0
  */
 
-static VALUE
+VALUE
 math_hypot(VALUE obj, VALUE x, VALUE y)
 {
     Need_Float2(x, y);
@@ -511,7 +566,7 @@ math_erfc(VALUE obj, VALUE x)
  *
  *   def fact(n) (1..n).inject(1) {|r,i| r*i } end
  *   0.upto(25) {|i| p [i, Math.gamma(i+1), fact(i)] }
- *   =>
+ *   #=>
  *   [0, 1.0, 1]
  *   [1, 1.0, 1]
  *   [2, 2.0, 2]
