@@ -2,17 +2,17 @@
 
   pack.c -
 
-  $Author: shyouhei $
-  $Date: 2009-02-19 18:01:35 +0900 (Thu, 19 Feb 2009) $
+  $Author: mame $
   created at: Thu Feb 10 15:17:05 JST 1994
 
-  Copyright (C) 1993-2003 Yukihiro Matsumoto
+  Copyright (C) 1993-2007 Yukihiro Matsumoto
 
 **********************************************************************/
 
-#include "ruby.h"
+#include "ruby/ruby.h"
 #include <sys/types.h>
 #include <ctype.h>
+#include <errno.h>
 
 #define SIZE16 2
 #define SIZE32 4
@@ -52,8 +52,7 @@
 
 #define define_swapx(x, xtype)		\
 static xtype				\
-TOKEN_PASTE(swap,x)(z)			\
-    xtype z;				\
+TOKEN_PASTE(swap,x)(xtype z)		\
 {					\
     xtype r;				\
     xtype *zp;				\
@@ -68,8 +67,8 @@ TOKEN_PASTE(swap,x)(z)			\
 	t[sizeof(xtype)-i-1] = s[i];	\
     }					\
     r = *(xtype *)t;			\
-    free(t);				\
-    free(zp);				\
+    xfree(t);				\
+    xfree(zp);				\
     return r;				\
 }
 
@@ -135,8 +134,7 @@ define_swapx(f,float)
 #else
 #if SIZEOF_LONG == 4	/* SIZEOF_DOUBLE == 8 && 4 == SIZEOF_LONG */
 static double
-swapd(d)
-    const double d;
+swapd(const double d)
 {
     double dtmp = d;
     unsigned long utmp[2];
@@ -153,8 +151,7 @@ swapd(d)
 #else
 #if SIZEOF_SHORT == 4	/* SIZEOF_DOUBLE == 8 && 4 == SIZEOF_SHORT */
 static double
-swapd(d)
-    const double d;
+swapd(const double d)
 {
     double dtmp = d;
     unsigned short utmp[2];
@@ -187,7 +184,7 @@ define_swapx(d, double)
 #undef htonl
 #endif
 static int
-endian()
+endian(void)
 {
     static int init = 0;
     static int endian_value;
@@ -331,11 +328,10 @@ endian()
 #define VTOHD(x,y)	vtohd(x)
 #endif
 
-unsigned long rb_big2ulong_pack _((VALUE x));
+unsigned long rb_big2ulong_pack(VALUE x);
 
 static unsigned long
-num2i32(x)
-    VALUE x;
+num2i32(VALUE x)
 {
     x = rb_to_int(x); /* is nil OK? (should not) */
 
@@ -348,13 +344,13 @@ num2i32(x)
 }
 
 #if SIZEOF_LONG == SIZE32
-# define EXTEND32(x) 
+# define EXTEND32(x)
 #else
 /* invariant in modulo 1<<31 */
 # define EXTEND32(x) do { if (!natint) {(x) = (((1L<<31)-1-(x))^~(~0L<<31));}} while(0)
 #endif
 #if SIZEOF_SHORT == SIZE16
-# define EXTEND16(x) 
+# define EXTEND16(x)
 #else
 # define EXTEND16(x) do { if (!natint) {(x) = (short)(((1<<15)-1-(x))^~(~0<<15));}} while(0)
 #endif
@@ -366,16 +362,15 @@ num2i32(x)
 #endif
 static const char toofew[] = "too few arguments";
 
-static void encodes _((VALUE,char*,long,int));
-static void qpencode _((VALUE,VALUE,long));
+static void encodes(VALUE,const char*,long,int);
+static void qpencode(VALUE,VALUE,long);
 
-static int uv_to_utf8 _((char*,unsigned long));
-static unsigned long utf8_to_uv _((char*,long*));
+static unsigned long utf8_to_uv(const char*,long*);
 
 /*
  *  call-seq:
  *     arr.pack ( aTemplateString ) -> aBinaryString
- *  
+ *
  *  Packs the contents of <i>arr</i> into a binary sequence according to
  *  the directives in <i>aTemplateString</i> (see the table below)
  *  Directives ``A,'' ``a,'' and ``Z'' may be followed by a count,
@@ -388,24 +383,24 @@ static unsigned long utf8_to_uv _((char*,long*));
  *  platform's native size for the specified type; otherwise, they use a
  *  platform-independent size. Spaces are ignored in the template
  *  string. See also <code>String#unpack</code>.
- *     
+ *
  *     a = [ "a", "b", "c" ]
  *     n = [ 65, 66, 67 ]
  *     a.pack("A3A3A3")   #=> "a  b  c  "
  *     a.pack("a3a3a3")   #=> "a\000\000b\000\000c\000\000"
  *     n.pack("ccc")      #=> "ABC"
- *     
+ *
  *  Directives for +pack+.
  *
  *   Directive    Meaning
  *   ---------------------------------------------------------------
  *       @     |  Moves to absolute position
- *       A     |  ASCII string (space padded, count is width)
- *       a     |  ASCII string (null padded, count is width)
+ *       A     |  arbitrary binary string (space padded, count is width)
+ *       a     |  arbitrary binary string (null padded, count is width)
  *       B     |  Bit string (descending bit order)
  *       b     |  Bit string (ascending bit order)
- *       C     |  Unsigned char
- *       c     |  Char
+ *       C     |  Unsigned byte (C unsigned char)
+ *       c     |  Byte (C char)
  *       D, d  |  Double-precision float, native format
  *       E     |  Double-precision float, little-endian byte order
  *       e     |  Single-precision float, little-endian byte order
@@ -438,34 +433,33 @@ static unsigned long utf8_to_uv _((char*,long*));
  */
 
 static VALUE
-pack_pack(ary, fmt)
-    VALUE ary, fmt;
+pack_pack(VALUE ary, VALUE fmt)
 {
-    static char *nul10 = "\0\0\0\0\0\0\0\0\0\0";
-    static char *spc10 = "          ";
-    char *p, *pend;
+    static const char nul10[] = "\0\0\0\0\0\0\0\0\0\0";
+    static const char spc10[] = "          ";
+    const char *p, *pend;
     VALUE res, from, associates = 0;
     char type;
     long items, len, idx, plen;
-    char *ptr;
+    const char *ptr;
 #ifdef NATINT_PACK
     int natint;		/* native integer */
 #endif
 
     StringValue(fmt);
-    p = RSTRING(fmt)->ptr;
-    pend = p + RSTRING(fmt)->len;
+    p = RSTRING_PTR(fmt);
+    pend = p + RSTRING_LEN(fmt);
     res = rb_str_buf_new(0);
 
-    items = RARRAY(ary)->len;
+    items = RARRAY_LEN(ary);
     idx = 0;
 
 #define TOO_FEW (rb_raise(rb_eArgError, toofew), 0)
-#define THISFROM (items > 0 ? RARRAY(ary)->ptr[idx] : TOO_FEW)
-#define NEXTFROM (items-- > 0 ? RARRAY(ary)->ptr[idx++] : TOO_FEW)
+#define THISFROM (items > 0 ? RARRAY_PTR(ary)[idx] : TOO_FEW)
+#define NEXTFROM (items-- > 0 ? RARRAY_PTR(ary)[idx++] : TOO_FEW)
 
     while (p < pend) {
-	if (RSTRING(fmt)->ptr + RSTRING(fmt)->len != pend) {
+	if (RSTRING_PTR(fmt) + RSTRING_LEN(fmt) != pend) {
 	    rb_raise(rb_eRuntimeError, "format string modified");
 	}
 	type = *p++;		/* get data type */
@@ -481,7 +475,7 @@ pack_pack(ary, fmt)
 	    continue;
 	}
         if (*p == '_' || *p == '!') {
-	    const char *natstr = "sSiIlL";
+	    static const char natstr[] = "sSiIlL";
 
 	    if (strchr(natstr, type)) {
 #ifdef NATINT_PACK
@@ -494,13 +488,15 @@ pack_pack(ary, fmt)
 	    }
 	}
 	if (*p == '*') {	/* set data length */
-	    len = strchr("@Xxu", type) ? 0
-                : strchr("PMm", type) ? 1
-                : items;
+	    len = strchr("@Xxu", type) ? 0 : items;
 	    p++;
 	}
 	else if (ISDIGIT(*p)) {
-	    len = strtoul(p, (char**)&p, 10);
+	    errno = 0;
+	    len = STRTOUL(p, (char**)&p, 10);
+	    if (errno) {
+		rb_raise(rb_eRangeError, "pack length too big");
+	    }
 	}
 	else {
 	    len = 1;
@@ -517,8 +513,8 @@ pack_pack(ary, fmt)
 	    }
 	    else {
 		StringValue(from);
-		ptr = RSTRING(from)->ptr;
-		plen = RSTRING(from)->len;
+		ptr = RSTRING_PTR(from);
+		plen = RSTRING_LEN(from);
 		OBJ_INFECT(res, from);
 	    }
 
@@ -527,8 +523,8 @@ pack_pack(ary, fmt)
 
 	    switch (type) {
 	      case 'a':		/* arbitrary binary string (null padded)  */
-	      case 'A':		/* ASCII string (space padded) */
-	      case 'Z':		/* null terminated ASCII string  */
+	      case 'A':         /* arbitrary binary string (ASCII space padded) */
+	      case 'Z':         /* null terminated string  */
 		if (plen >= len) {
 		    rb_str_buf_cat(res, ptr, len);
 		    if (p[-1] == '*' && type == 'Z')
@@ -612,7 +608,7 @@ pack_pack(ary, fmt)
 		    long i, j = 0;
 
 		    if (len > plen) {
-			j = (len + 1) / 2 - (plen + 1) / 2;
+			j = (len - plen + 1)/2;
 			len = plen;
 		    }
 		    for (i=0; i++ < len; ptr++) {
@@ -643,7 +639,7 @@ pack_pack(ary, fmt)
 		    long i, j = 0;
 
 		    if (len > plen) {
-			j = (len + 1) / 2 - (plen + 1) / 2;
+			j = (len - plen + 1)/2;
 			len = plen;
 		    }
 		    for (i=0; i++ < len; ptr++) {
@@ -695,11 +691,11 @@ pack_pack(ary, fmt)
 	  case 'i':		/* signed int */
 	  case 'I':		/* unsigned int */
 	    while (len-- > 0) {
-		long i;
+		int i;
 
 		from = NEXTFROM;
 		i = num2i32(from);
-		rb_str_buf_cat(res, OFF32(&i), NATINT_LEN(int,4));
+		rb_str_buf_cat(res, (char*)&i, sizeof(int));
 	    }
 	    break;
 
@@ -775,7 +771,7 @@ pack_pack(ary, fmt)
 		float f;
 
 		from = NEXTFROM;
-		f = RFLOAT(rb_Float(from))->value;
+		f = RFLOAT_VALUE(rb_Float(from));
 		rb_str_buf_cat(res, (char*)&f, sizeof(float));
 	    }
 	    break;
@@ -786,7 +782,7 @@ pack_pack(ary, fmt)
 		FLOAT_CONVWITH(ftmp);
 
 		from = NEXTFROM;
-		f = RFLOAT(rb_Float(from))->value;
+		f = RFLOAT_VALUE(rb_Float(from));
 		f = HTOVF(f,ftmp);
 		rb_str_buf_cat(res, (char*)&f, sizeof(float));
 	    }
@@ -798,7 +794,7 @@ pack_pack(ary, fmt)
 		DOUBLE_CONVWITH(dtmp);
 
 		from = NEXTFROM;
-		d = RFLOAT(rb_Float(from))->value;
+		d = RFLOAT_VALUE(rb_Float(from));
 		d = HTOVD(d,dtmp);
 		rb_str_buf_cat(res, (char*)&d, sizeof(double));
 	    }
@@ -810,7 +806,7 @@ pack_pack(ary, fmt)
 		double d;
 
 		from = NEXTFROM;
-		d = RFLOAT(rb_Float(from))->value;
+		d = RFLOAT_VALUE(rb_Float(from));
 		rb_str_buf_cat(res, (char*)&d, sizeof(double));
 	    }
 	    break;
@@ -821,7 +817,7 @@ pack_pack(ary, fmt)
 		FLOAT_CONVWITH(ftmp);
 
 		from = NEXTFROM;
-		f = RFLOAT(rb_Float(from))->value;
+		f = RFLOAT_VALUE(rb_Float(from));
 		f = HTONF(f,ftmp);
 		rb_str_buf_cat(res, (char*)&f, sizeof(float));
 	    }
@@ -833,7 +829,7 @@ pack_pack(ary, fmt)
 		DOUBLE_CONVWITH(dtmp);
 
 		from = NEXTFROM;
-		d = RFLOAT(rb_Float(from))->value;
+		d = RFLOAT_VALUE(rb_Float(from));
 		d = HTOND(d,dtmp);
 		rb_str_buf_cat(res, (char*)&d, sizeof(double));
 	    }
@@ -850,15 +846,14 @@ pack_pack(ary, fmt)
 
 	  case 'X':		/* back up byte */
 	  shrink:
-	    plen = RSTRING(res)->len;
+	    plen = RSTRING_LEN(res);
 	    if (plen < len)
 		rb_raise(rb_eArgError, "X outside of string");
-	    RSTRING(res)->len = plen - len;
-	    RSTRING(res)->ptr[plen - len] = '\0';
+	    rb_str_set_len(res, plen - len);
 	    break;
 
 	  case '@':		/* null fill to absolute position */
-	    len -= RSTRING(res)->len;
+	    len -= RSTRING_LEN(res);
 	    if (len > 0) goto grow;
 	    len = -len;
 	    if (len > 0) goto shrink;
@@ -870,17 +865,17 @@ pack_pack(ary, fmt)
 
 	  case 'U':		/* Unicode character */
 	    while (len-- > 0) {
-		long l;
+		SIGNED_VALUE l;
 		char buf[8];
 		int le;
 
 		from = NEXTFROM;
 		from = rb_to_int(from);
-		l = NUM2INT(from);
+		l = NUM2LONG(from);
 		if (l < 0) {
 		    rb_raise(rb_eRangeError, "pack(U): value out of range");
 		}
-		le = uv_to_utf8(buf, l);
+		le = rb_uv_to_utf8(buf, l);
 		rb_str_buf_cat(res, (char*)buf, le);
 	    }
 	    break;
@@ -889,8 +884,8 @@ pack_pack(ary, fmt)
 	  case 'm':		/* base64 encoded string */
 	    from = NEXTFROM;
 	    StringValue(from);
-	    ptr = RSTRING(from)->ptr;
-	    plen = RSTRING(from)->len;
+	    ptr = RSTRING_PTR(from);
+	    plen = RSTRING_LEN(from);
 
 	    if (len <= 2)
 		len = 45;
@@ -920,9 +915,9 @@ pack_pack(ary, fmt)
 	    from = THISFROM;
 	    if (!NIL_P(from)) {
 		StringValue(from);
-		if (RSTRING(from)->len < len) {
+		if (RSTRING_LEN(from) < len) {
 		    rb_raise(rb_eArgError, "too short buffer for P(%ld for %ld)",
-			     RSTRING(from)->len, len);
+			     RSTRING_LEN(from), len);
 		}
 	    }
 	    len = 1;
@@ -957,9 +952,9 @@ pack_pack(ary, fmt)
 		    VALUE big128 = rb_uint2big(128);
 		    while (TYPE(from) == T_BIGNUM) {
 			from = rb_big_divmod(from, big128);
-			c = NUM2INT(RARRAY(from)->ptr[1]) | 0x80; /* mod */
+			c = NUM2INT(RARRAY_PTR(from)[1]) | 0x80; /* mod */
 			rb_str_buf_cat(buf, &c, sizeof(char));
-			from = RARRAY(from)->ptr[0]; /* div */
+			from = RARRAY_PTR(from)[0]; /* div */
 		    }
 		}
 
@@ -977,16 +972,16 @@ pack_pack(ary, fmt)
 		    ul >>=  7;
 		}
 
-		if (RSTRING(buf)->len) {
-		    bufs = RSTRING(buf)->ptr;
-		    bufe = bufs + RSTRING(buf)->len - 1;
+		if (RSTRING_LEN(buf)) {
+		    bufs = RSTRING_PTR(buf);
+		    bufe = bufs + RSTRING_LEN(buf) - 1;
 		    *bufs &= 0x7f; /* clear continue bit */
 		    while (bufs < bufe) { /* reverse */
 			c = *bufs;
 			*bufs++ = *bufe;
 			*bufe-- = c;
 		    }
-		    rb_str_buf_cat(res, RSTRING(buf)->ptr, RSTRING(buf)->len);
+		    rb_str_buf_cat(res, RSTRING_PTR(buf), RSTRING_LEN(buf));
 		}
 		else {
 		    c = 0;
@@ -1003,25 +998,20 @@ pack_pack(ary, fmt)
     if (associates) {
 	rb_str_associate(res, associates);
     }
-    OBJ_INFECT(res, fmt);
     return res;
 }
 
-static char uu_table[] =
+static const char uu_table[] =
 "`!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_";
-static char b64_table[] =
+static const char b64_table[] =
 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 
 static void
-encodes(str, s, len, type)
-    VALUE str;
-    char *s;
-    long len;
-    int type;
+encodes(VALUE str, const char *s, long len, int type)
 {
     char *buff = ALLOCA_N(char, len * 4 / 3 + 6);
     long i = 0;
-    char *trans = type == 'u' ? uu_table : b64_table;
+    const char *trans = type == 'u' ? uu_table : b64_table;
     int padding;
 
     if (type == 'u') {
@@ -1055,17 +1045,15 @@ encodes(str, s, len, type)
     rb_str_buf_cat(str, buff, i);
 }
 
-static char hex_table[] = "0123456789ABCDEF";
+static const char hex_table[] = "0123456789ABCDEF";
 
 static void
-qpencode(str, from, len)
-    VALUE str, from;
-    long len;
+qpencode(VALUE str, VALUE from, long len)
 {
     char buff[1024];
     long i = 0, n = 0, prev = EOF;
-    unsigned char *s = (unsigned char*)RSTRING(from)->ptr;
-    unsigned char *send = s + RSTRING(from)->len;
+    unsigned char *s = (unsigned char*)RSTRING_PTR(from);
+    unsigned char *send = s + RSTRING_LEN(from);
 
     while (s < send) {
         if ((*s > 126) ||
@@ -1113,20 +1101,19 @@ qpencode(str, from, len)
 }
 
 static inline int
-hex2num(c)
-    char c;
+hex2num(char c)
 {
     switch (c) {
-    case '0': case '1': case '2': case '3': case '4':
-    case '5': case '6': case '7': case '8': case '9':
+      case '0': case '1': case '2': case '3': case '4':
+      case '5': case '6': case '7': case '8': case '9':
         return c - '0';
-    case 'a': case 'b': case 'c':
-    case 'd': case 'e': case 'f':
+      case 'a': case 'b': case 'c':
+      case 'd': case 'e': case 'f':
 	return c - 'a' + 10;
-    case 'A': case 'B': case 'C':
-    case 'D': case 'E': case 'F':
+      case 'A': case 'B': case 'C':
+      case 'D': case 'E': case 'F':
 	return c - 'A' + 10;
-    default:
+      default:
 	return -1;
     }
 }
@@ -1154,21 +1141,18 @@ hex2num(c)
 #define PACK_ITEM_ADJUST() while (tmp--) rb_ary_push(ary, Qnil)
 
 static VALUE
-infected_str_new(ptr, len, str)
-    const char *ptr;
-    long len;
-    VALUE str;
+infected_str_new(const char *ptr, long len, VALUE str)
 {
     VALUE s = rb_str_new(ptr, len);
 
     OBJ_INFECT(s, str);
     return s;
 }
-    
+
 /*
  *  call-seq:
  *     str.unpack(format)   => anArray
- *  
+ *
  *  Decodes <i>str</i> (which may contain binary data) according to the
  *  format string, returning an array of each value extracted. The
  *  format string consists of a sequence of single-character directives,
@@ -1181,7 +1165,7 @@ infected_str_new(ptr, len, str)
  *  platform's native size for the specified type; otherwise, it uses a
  *  platform-independent consistent size. Spaces are ignored in the
  *  format string. See also <code>Array#pack</code>.
- *     
+ *
  *     "abc \0\0abc \0\0".unpack('A6Z6')   #=> ["abc", "abc "]
  *     "abc \0\0".unpack('a3a3')           #=> ["abc", " \000\000"]
  *     "abc \0abc \0".unpack('Z*Z*')       #=> ["abc ", "abc "]
@@ -1193,20 +1177,21 @@ infected_str_new(ptr, len, str)
  *
  *  This table summarizes the various formats and the Ruby classes
  *  returned by each.
- *     
+ *
  *     Format | Returns | Function
  *     -------+---------+-----------------------------------------
- *       A    | String  | with trailing nulls and spaces removed
+ *       A    | String  | arbitrary binary string with trailing
+ *            |         | nulls and ASCII spaces removed
  *     -------+---------+-----------------------------------------
- *       a    | String  | string
+ *       a    | String  | arbitrary binary string
  *     -------+---------+-----------------------------------------
  *       B    | String  | extract bits from each character (msb first)
  *     -------+---------+-----------------------------------------
  *       b    | String  | extract bits from each character (lsb first)
  *     -------+---------+-----------------------------------------
- *       C    | Fixnum  | extract a character as an unsigned integer
+ *       C    | Fixnum  | extract a byte (C char) as an unsigned integer
  *     -------+---------+-----------------------------------------
- *       c    | Fixnum  | extract a character as an integer
+ *       c    | Fixnum  | extract a byte (C char) as an integer
  *     -------+---------+-----------------------------------------
  *       d,D  | Float   | treat sizeof(double) characters as
  *            |         | a native double
@@ -1265,17 +1250,17 @@ infected_str_new(ptr, len, str)
  *       p    | String  | treat sizeof(char *) characters as a
  *            |         | pointer to a  null-terminated string
  *     -------+---------+-----------------------------------------
- *       Q    | Integer | treat 8 characters as an unsigned 
+ *       Q    | Integer | treat 8 characters as an unsigned
  *            |         | quad word (64 bits)
  *     -------+---------+-----------------------------------------
- *       q    | Integer | treat 8 characters as a signed 
+ *       q    | Integer | treat 8 characters as a signed
  *            |         | quad word (64 bits)
  *     -------+---------+-----------------------------------------
  *       S    | Fixnum  | treat two (different if _ used)
  *            |         | successive characters as an unsigned
  *            |         | short in native byte order
  *     -------+---------+-----------------------------------------
- *       s    | Fixnum  | Treat two (different if _ used) 
+ *       s    | Fixnum  | Treat two (different if _ used)
  *            |         | successive characters as a signed short
  *            |         | in native byte order
  *     -------+---------+-----------------------------------------
@@ -1298,16 +1283,15 @@ infected_str_new(ptr, len, str)
  *       Z    | String  | with trailing nulls removed
  *            |         | upto first null with *
  *     -------+---------+-----------------------------------------
- *       @    | ---     | skip to the offset given by the 
+ *       @    | ---     | skip to the offset given by the
  *            |         | length argument
  *     -------+---------+-----------------------------------------
  */
 
 static VALUE
-pack_unpack(str, fmt)
-    VALUE str, fmt;
+pack_unpack(VALUE str, VALUE fmt)
 {
-    static char *hexdigits = "0123456789abcdef0123456789ABCDEFx";
+    static const char *hexdigits = "0123456789abcdef0123456789ABCDEFx";
     char *s, *send;
     char *p, *pend;
     VALUE ary;
@@ -1317,15 +1301,25 @@ pack_unpack(str, fmt)
 #ifdef NATINT_PACK
     int natint;			/* native integer */
 #endif
+    int block_p = rb_block_given_p();
+#define UNPACK_PUSH(item) do {\
+	VALUE item_val = (item);\
+	if (block_p) {\
+	    rb_yield(item_val);\
+	}\
+	else {\
+	    rb_ary_push(ary, item_val);\
+	}\
+    } while (0)
 
     StringValue(str);
     StringValue(fmt);
-    s = RSTRING(str)->ptr;
-    send = s + RSTRING(str)->len;
-    p = RSTRING(fmt)->ptr;
-    pend = p + RSTRING(fmt)->len;
+    s = RSTRING_PTR(str);
+    send = s + RSTRING_LEN(str);
+    p = RSTRING_PTR(fmt);
+    pend = p + RSTRING_LEN(fmt);
 
-    ary = rb_ary_new();
+    ary = block_p ? Qnil : rb_ary_new();
     while (p < pend) {
 	type = *p++;
 #ifdef NATINT_PACK
@@ -1341,7 +1335,7 @@ pack_unpack(str, fmt)
 	}
 	star = 0;
 	if (*p == '_' || *p == '!') {
-	    char *natstr = "sSiIlL";
+	    static const char natstr[] = "sSiIlL";
 
 	    if (strchr(natstr, type)) {
 #ifdef NATINT_PACK
@@ -1361,7 +1355,11 @@ pack_unpack(str, fmt)
 	    p++;
 	}
 	else if (ISDIGIT(*p)) {
-	    len = strtoul(p, (char**)&p, 10);
+	    errno = 0;
+	    len = STRTOUL(p, (char**)&p, 10);
+	    if (errno) {
+		rb_raise(rb_eRangeError, "pack length too big");
+	    }
 	}
 	else {
 	    len = (type != '@');
@@ -1382,7 +1380,7 @@ pack_unpack(str, fmt)
 		    if (*t != ' ' && *t != '\0') break;
 		    t--; len--;
 		}
-		rb_ary_push(ary, infected_str_new(s, len, str));
+		UNPACK_PUSH(infected_str_new(s, len, str));
 		s += end;
 	    }
 	    break;
@@ -1393,7 +1391,7 @@ pack_unpack(str, fmt)
 
 		if (len > send-s) len = send-s;
 		while (t < s+len && *t) t++;
-		rb_ary_push(ary, infected_str_new(s, t-s, str));
+		UNPACK_PUSH(infected_str_new(s, t-s, str));
 		if (t < send) t++;
 		s = star ? t : s+len;
 	    }
@@ -1401,7 +1399,7 @@ pack_unpack(str, fmt)
 
 	  case 'a':
 	    if (len > send - s) len = send - s;
-	    rb_ary_push(ary, infected_str_new(s, len, str));
+	    UNPACK_PUSH(infected_str_new(s, len, str));
 	    s += len;
 	    break;
 
@@ -1415,8 +1413,8 @@ pack_unpack(str, fmt)
 		if (p[-1] == '*' || len > (send - s) * 8)
 		    len = (send - s) * 8;
 		bits = 0;
-		rb_ary_push(ary, bitstr = rb_str_new(0, len));
-		t = RSTRING(bitstr)->ptr;
+		UNPACK_PUSH(bitstr = rb_str_new(0, len));
+		t = RSTRING_PTR(bitstr);
 		for (i=0; i<len; i++) {
 		    if (i & 7) bits >>= 1;
 		    else bits = *s++;
@@ -1435,8 +1433,8 @@ pack_unpack(str, fmt)
 		if (p[-1] == '*' || len > (send - s) * 8)
 		    len = (send - s) * 8;
 		bits = 0;
-		rb_ary_push(ary, bitstr = rb_str_new(0, len));
-		t = RSTRING(bitstr)->ptr;
+		UNPACK_PUSH(bitstr = rb_str_new(0, len));
+		t = RSTRING_PTR(bitstr);
 		for (i=0; i<len; i++) {
 		    if (i & 7) bits <<= 1;
 		    else bits = *s++;
@@ -1455,8 +1453,8 @@ pack_unpack(str, fmt)
 		if (p[-1] == '*' || len > (send - s) * 2)
 		    len = (send - s) * 2;
 		bits = 0;
-		rb_ary_push(ary, bitstr = rb_str_new(0, len));
-		t = RSTRING(bitstr)->ptr;
+		UNPACK_PUSH(bitstr = rb_str_new(0, len));
+		t = RSTRING_PTR(bitstr);
 		for (i=0; i<len; i++) {
 		    if (i & 1)
 			bits >>= 4;
@@ -1477,8 +1475,8 @@ pack_unpack(str, fmt)
 		if (p[-1] == '*' || len > (send - s) * 2)
 		    len = (send - s) * 2;
 		bits = 0;
-		rb_ary_push(ary, bitstr = rb_str_new(0, len));
-		t = RSTRING(bitstr)->ptr;
+		UNPACK_PUSH(bitstr = rb_str_new(0, len));
+		t = RSTRING_PTR(bitstr);
 		for (i=0; i<len; i++) {
 		    if (i & 1)
 			bits <<= 4;
@@ -1494,7 +1492,7 @@ pack_unpack(str, fmt)
 	    while (len-- > 0) {
                 int c = *s++;
                 if (c > (char)127) c-=256;
-		rb_ary_push(ary, INT2FIX(c));
+		UNPACK_PUSH(INT2FIX(c));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1503,7 +1501,7 @@ pack_unpack(str, fmt)
 	    PACK_LENGTH_ADJUST(unsigned char,sizeof(unsigned char));
 	    while (len-- > 0) {
 		unsigned char c = *s++;
-		rb_ary_push(ary, INT2FIX(c));
+		UNPACK_PUSH(INT2FIX(c));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1515,7 +1513,7 @@ pack_unpack(str, fmt)
 		memcpy(OFF16(&tmp), s, NATINT_LEN(short,2));
 		EXTEND16(tmp);
 		s += NATINT_LEN(short,2);
-		rb_ary_push(ary, INT2FIX(tmp));
+		UNPACK_PUSH(INT2FIX(tmp));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1526,7 +1524,7 @@ pack_unpack(str, fmt)
 		unsigned short tmp = 0;
 		memcpy(OFF16(&tmp), s, NATINT_LEN(unsigned short,2));
 		s += NATINT_LEN(unsigned short,2);
-		rb_ary_push(ary, INT2FIX(tmp));
+		UNPACK_PUSH(INT2FIX(tmp));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1537,7 +1535,7 @@ pack_unpack(str, fmt)
 		int tmp;
 		memcpy(&tmp, s, sizeof(int));
 		s += sizeof(int);
-		rb_ary_push(ary, INT2NUM(tmp));
+		UNPACK_PUSH(INT2NUM(tmp));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1548,7 +1546,7 @@ pack_unpack(str, fmt)
 		unsigned int tmp;
 		memcpy(&tmp, s, sizeof(unsigned int));
 		s += sizeof(unsigned int);
-		rb_ary_push(ary, UINT2NUM(tmp));
+		UNPACK_PUSH(UINT2NUM(tmp));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1560,18 +1558,17 @@ pack_unpack(str, fmt)
 		memcpy(OFF32(&tmp), s, NATINT_LEN(long,4));
 		EXTEND32(tmp);
 		s += NATINT_LEN(long,4);
-		rb_ary_push(ary, LONG2NUM(tmp));
+		UNPACK_PUSH(LONG2NUM(tmp));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
-
 	  case 'L':
 	    PACK_LENGTH_ADJUST(unsigned long,4);
 	    while (len-- > 0) {
 		unsigned long tmp = 0;
 		memcpy(OFF32(&tmp), s, NATINT_LEN(unsigned long,4));
 		s += NATINT_LEN(unsigned long,4);
-		rb_ary_push(ary, ULONG2NUM(tmp));
+		UNPACK_PUSH(ULONG2NUM(tmp));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1581,17 +1578,16 @@ pack_unpack(str, fmt)
 	    while (len-- > 0) {
 		char *tmp = (char*)s;
 		s += QUAD_SIZE;
-		rb_ary_push(ary, rb_quad_unpack(tmp, 1));
+		UNPACK_PUSH(rb_quad_unpack(tmp, 1));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
-
 	  case 'Q':
 	    PACK_LENGTH_ADJUST_SIZE(QUAD_SIZE);
 	    while (len-- > 0) {
 		char *tmp = (char*)s;
 		s += QUAD_SIZE;
-		rb_ary_push(ary, rb_quad_unpack(tmp, 0));
+		UNPACK_PUSH(rb_quad_unpack(tmp, 0));
 	    }
 	    break;
 
@@ -1601,7 +1597,7 @@ pack_unpack(str, fmt)
 		unsigned short tmp = 0;
 		memcpy(OFF16B(&tmp), s, NATINT_LEN(unsigned short,2));
 		s += NATINT_LEN(unsigned short,2);
-		rb_ary_push(ary, UINT2NUM(ntohs(tmp)));
+		UNPACK_PUSH(UINT2NUM(ntohs(tmp)));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1612,7 +1608,7 @@ pack_unpack(str, fmt)
 		unsigned long tmp = 0;
 		memcpy(OFF32B(&tmp), s, NATINT_LEN(unsigned long,4));
 		s += NATINT_LEN(unsigned long,4);
-		rb_ary_push(ary, ULONG2NUM(ntohl(tmp)));
+		UNPACK_PUSH(ULONG2NUM(ntohl(tmp)));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1621,9 +1617,9 @@ pack_unpack(str, fmt)
 	    PACK_LENGTH_ADJUST(unsigned short,2);
 	    while (len-- > 0) {
 		unsigned short tmp = 0;
-		memcpy(OFF16(&tmp), s, NATINT_LEN(unsigned short,2));
+		memcpy(&tmp, s, NATINT_LEN(unsigned short,2));
 		s += NATINT_LEN(unsigned short,2);
-		rb_ary_push(ary, UINT2NUM(vtohs(tmp)));
+		UNPACK_PUSH(UINT2NUM(vtohs(tmp)));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1632,9 +1628,9 @@ pack_unpack(str, fmt)
 	    PACK_LENGTH_ADJUST(unsigned long,4);
 	    while (len-- > 0) {
 		unsigned long tmp = 0;
-		memcpy(OFF32(&tmp), s, NATINT_LEN(long,4));
+		memcpy(&tmp, s, NATINT_LEN(long,4));
 		s += NATINT_LEN(long,4);
-		rb_ary_push(ary, ULONG2NUM(vtohl(tmp)));
+		UNPACK_PUSH(ULONG2NUM(vtohl(tmp)));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1646,7 +1642,7 @@ pack_unpack(str, fmt)
 		float tmp;
 		memcpy(&tmp, s, sizeof(float));
 		s += sizeof(float);
-		rb_ary_push(ary, rb_float_new((double)tmp));
+		UNPACK_PUSH(DOUBLE2NUM((double)tmp));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1660,11 +1656,11 @@ pack_unpack(str, fmt)
 		memcpy(&tmp, s, sizeof(float));
 		s += sizeof(float);
 		tmp = VTOHF(tmp,ftmp);
-		rb_ary_push(ary, rb_float_new((double)tmp));
+		UNPACK_PUSH(DOUBLE2NUM((double)tmp));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
-	    
+
 	  case 'E':
 	    PACK_LENGTH_ADJUST(double,sizeof(double));
 	    while (len-- > 0) {
@@ -1674,11 +1670,11 @@ pack_unpack(str, fmt)
 		memcpy(&tmp, s, sizeof(double));
 		s += sizeof(double);
 		tmp = VTOHD(tmp,dtmp);
-		rb_ary_push(ary, rb_float_new(tmp));
+		UNPACK_PUSH(DOUBLE2NUM(tmp));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
-	    
+
 	  case 'D':
 	  case 'd':
 	    PACK_LENGTH_ADJUST(double,sizeof(double));
@@ -1686,7 +1682,7 @@ pack_unpack(str, fmt)
 		double tmp;
 		memcpy(&tmp, s, sizeof(double));
 		s += sizeof(double);
-		rb_ary_push(ary, rb_float_new(tmp));
+		UNPACK_PUSH(DOUBLE2NUM(tmp));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
@@ -1700,11 +1696,11 @@ pack_unpack(str, fmt)
 		memcpy(&tmp, s, sizeof(float));
 		s += sizeof(float);
 		tmp = NTOHF(tmp,ftmp);
-		rb_ary_push(ary, rb_float_new((double)tmp));
+		UNPACK_PUSH(DOUBLE2NUM((double)tmp));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
-	    
+
 	  case 'G':
 	    PACK_LENGTH_ADJUST(double,sizeof(double));
 	    while (len-- > 0) {
@@ -1714,11 +1710,11 @@ pack_unpack(str, fmt)
 		memcpy(&tmp, s, sizeof(double));
 		s += sizeof(double);
 		tmp = NTOHD(tmp,dtmp);
-		rb_ary_push(ary, rb_float_new(tmp));
+		UNPACK_PUSH(DOUBLE2NUM(tmp));
 	    }
 	    PACK_ITEM_ADJUST();
 	    break;
-	    
+
 	  case 'U':
 	    if (len > send - s) len = send - s;
 	    while (len > 0 && s < send) {
@@ -1727,14 +1723,14 @@ pack_unpack(str, fmt)
 
 		l = utf8_to_uv(s, &alen);
 		s += alen; len--;
-		rb_ary_push(ary, ULONG2NUM(l));
+		UNPACK_PUSH(ULONG2NUM(l));
 	    }
 	    break;
 
 	  case 'u':
 	    {
 		VALUE buf = infected_str_new(0, (send - s)*3/4, str);
-		char *ptr = RSTRING(buf)->ptr;
+		char *ptr = RSTRING_PTR(buf);
 		long total = 0;
 
 		while (s < send && *s > ' ' && *s < 'a') {
@@ -1744,9 +1740,9 @@ pack_unpack(str, fmt)
 		    hunk[3] = '\0';
 		    len = (*s++ - ' ') & 077;
 		    total += len;
-		    if (total > RSTRING(buf)->len) {
-			len -= total - RSTRING(buf)->len;
-			total = RSTRING(buf)->len;
+		    if (total > RSTRING_LEN(buf)) {
+			len -= total - RSTRING_LEN(buf);
+			total = RSTRING_LEN(buf);
 		    }
 
 		    while (len > 0) {
@@ -1780,17 +1776,16 @@ pack_unpack(str, fmt)
 		    else if (s < send && (s+1 == send || s[1] == '\n'))
 			s += 2;	/* possible checksum byte */
 		}
-		
-		RSTRING(buf)->ptr[total] = '\0';
-		RSTRING(buf)->len = total;
-		rb_ary_push(ary, buf);
+
+		rb_str_set_len(buf, total);
+		UNPACK_PUSH(buf);
 	    }
 	    break;
 
 	  case 'm':
 	    {
 		VALUE buf = infected_str_new(0, (send - s)*3/4, str);
-		char *ptr = RSTRING(buf)->ptr;
+		char *ptr = RSTRING_PTR(buf);
 		int a = -1,b = -1,c = 0,d;
 		static int first = 1;
 		static int b64_xtable[256];
@@ -1832,16 +1827,15 @@ pack_unpack(str, fmt)
 			*ptr++ = b << 4 | c >> 2;
 		    }
 		}
-		*ptr = '\0';
-		RSTRING(buf)->len = ptr - RSTRING(buf)->ptr;
-		rb_ary_push(ary, buf);
+		rb_str_set_len(buf, ptr - RSTRING_PTR(buf));
+		UNPACK_PUSH(buf);
 	    }
 	    break;
 
 	  case 'M':
 	    {
 		VALUE buf = infected_str_new(0, send - s, str);
-		char *ptr = RSTRING(buf)->ptr;
+		char *ptr = RSTRING_PTR(buf);
 		int c1, c2;
 
 		while (s < send) {
@@ -1861,20 +1855,19 @@ pack_unpack(str, fmt)
 		    }
 		    s++;
 		}
-		*ptr = '\0';
-		RSTRING(buf)->len = ptr - RSTRING(buf)->ptr;
-		rb_ary_push(ary, buf);
+		rb_str_set_len(buf, ptr - RSTRING_PTR(buf));
+		UNPACK_PUSH(buf);
 	    }
 	    break;
 
 	  case '@':
-	    if (len > RSTRING(str)->len)
+	    if (len > RSTRING_LEN(str))
 		rb_raise(rb_eArgError, "@ outside of string");
-	    s = RSTRING(str)->ptr + len;
+	    s = RSTRING_PTR(str) + len;
 	    break;
 
 	  case 'X':
-	    if (len > s - RSTRING(str)->ptr)
+	    if (len > s - RSTRING_PTR(str))
 		rb_raise(rb_eArgError, "X outside of string");
 	    s -= len;
 	    break;
@@ -1887,8 +1880,8 @@ pack_unpack(str, fmt)
 
 	  case 'P':
 	    if (sizeof(char *) <= send - s) {
+		VALUE tmp = Qnil;
 		char *t;
-		VALUE tmp;
 
 		memcpy(&t, s, sizeof(char *));
 		s += sizeof(char *);
@@ -1899,11 +1892,11 @@ pack_unpack(str, fmt)
 		    if (!(a = rb_str_associated(str))) {
 			rb_raise(rb_eArgError, "no associated pointer");
 		    }
-		    p = RARRAY(a)->ptr;
-		    pend = p + RARRAY(a)->len;
+		    p = RARRAY_PTR(a);
+		    pend = p + RARRAY_LEN(a);
 		    while (p < pend) {
-			if (TYPE(*p) == T_STRING && RSTRING(*p)->ptr == t) {
-			    if (len < RSTRING(*p)->len) {
+			if (TYPE(*p) == T_STRING && RSTRING_PTR(*p) == t) {
+			    if (len < RSTRING_LEN(*p)) {
 				tmp = rb_tainted_str_new(t, len);
 				rb_str_associate(tmp, a);
 			    }
@@ -1918,10 +1911,7 @@ pack_unpack(str, fmt)
 			rb_raise(rb_eArgError, "non associated pointer");
 		    }
 		}
-		else {
-		    tmp = Qnil;
-		}
-		rb_ary_push(ary, tmp);
+		UNPACK_PUSH(tmp);
 	    }
 	    break;
 
@@ -1932,7 +1922,7 @@ pack_unpack(str, fmt)
 		if (send - s < sizeof(char *))
 		    break;
 		else {
-		    VALUE tmp;
+		    VALUE tmp = Qnil;
 		    char *t;
 
 		    memcpy(&t, s, sizeof(char *));
@@ -1944,10 +1934,10 @@ pack_unpack(str, fmt)
 			if (!(a = rb_str_associated(str))) {
 			    rb_raise(rb_eArgError, "no associated pointer");
 			}
-			p = RARRAY(a)->ptr;
-			pend = p + RARRAY(a)->len;
+			p = RARRAY_PTR(a);
+			pend = p + RARRAY_LEN(a);
 			while (p < pend) {
-			    if (TYPE(*p) == T_STRING && RSTRING(*p)->ptr == t) {
+			    if (TYPE(*p) == T_STRING && RSTRING_PTR(*p) == t) {
 				tmp = *p;
 				break;
 			    }
@@ -1957,10 +1947,7 @@ pack_unpack(str, fmt)
 			    rb_raise(rb_eArgError, "non associated pointer");
 			}
 		    }
-		    else {
-			tmp = Qnil;
-		    }
-		    rb_ary_push(ary, tmp);
+		    UNPACK_PUSH(tmp);
 		}
 	    }
 	    break;
@@ -1974,7 +1961,7 @@ pack_unpack(str, fmt)
 		    ul <<= 7;
 		    ul |= (*s & 0x7f);
 		    if (!(*s++ & 0x80)) {
-			rb_ary_push(ary, ULONG2NUM(ul));
+			UNPACK_PUSH(ULONG2NUM(ul));
 			len--;
 			ul = 0;
 		    }
@@ -1985,7 +1972,7 @@ pack_unpack(str, fmt)
 			    big = rb_big_mul(big, big128);
 			    big = rb_big_plus(big, rb_uint2big(*s & 0x7f));
 			    if (!(*s++ & 0x80)) {
-				rb_ary_push(ary, big);
+				UNPACK_PUSH(big);
 				len--;
 				ul = 0;
 				break;
@@ -2006,10 +1993,8 @@ pack_unpack(str, fmt)
 
 #define BYTEWIDTH 8
 
-static int
-uv_to_utf8(buf, uv)
-    char *buf;
-    unsigned long uv;
+int
+rb_uv_to_utf8(char buf[6], unsigned long uv)
 {
     if (uv <= 0x7f) {
 	buf[0] = (char)uv;
@@ -2064,9 +2049,7 @@ static const unsigned long utf8_limits[] = {
 };
 
 static unsigned long
-utf8_to_uv(p, lenp)
-    char *p;
-    long *lenp;
+utf8_to_uv(const char *p, long *lenp)
 {
     int c = *p++ & 0xff;
     unsigned long uv = c;
@@ -2091,7 +2074,7 @@ utf8_to_uv(p, lenp)
 	rb_raise(rb_eArgError, "malformed UTF-8 character");
     }
     if (n > *lenp) {
-	rb_raise(rb_eArgError, "malformed UTF-8 character (expected %d bytes, given %d bytes)",
+	rb_raise(rb_eArgError, "malformed UTF-8 character (expected %ld bytes, given %ld bytes)",
 		 n, *lenp);
     }
     *lenp = n--;
@@ -2116,7 +2099,7 @@ utf8_to_uv(p, lenp)
 }
 
 void
-Init_pack()
+Init_pack(void)
 {
     rb_define_method(rb_cArray, "pack", pack_pack, 1);
     rb_define_method(rb_cString, "unpack", pack_unpack, 1);
