@@ -2,7 +2,7 @@
 
   time.c -
 
-  $Author: matz $
+  $Author: nobu $
   created at: Tue Dec 28 14:31:59 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -12,6 +12,7 @@
 #include "ruby/ruby.h"
 #include <sys/types.h>
 #include <time.h>
+#include <errno.h>
 #include "ruby/encoding.h"
 
 #ifdef HAVE_UNISTD_H
@@ -38,7 +39,7 @@ struct time_object {
 static void
 time_free(void *tobj)
 {
-    if (tobj) free(tobj);
+    if (tobj) xfree(tobj);
 }
 
 static VALUE
@@ -200,6 +201,10 @@ time_timespec(VALUE num, int interval)
 	    double f, d;
 
 	    d = modf(RFLOAT_VALUE(num), &f);
+            if (d < 0) {
+                d += 1;
+                f -= 1;
+            }
 	    t.tv_sec = (time_t)f;
 	    if (f != t.tv_sec) {
 		rb_raise(rb_eRangeError, "%f out of Time range", RFLOAT_VALUE(num));
@@ -216,18 +221,24 @@ time_timespec(VALUE num, int interval)
 	break;
 
       default:
-        ary = rb_check_array_type(rb_funcall(num, id_divmod, 1, INT2FIX(1)));
-        if (NIL_P(ary)) {
+        if (rb_respond_to(num, id_divmod)) {
+            ary = rb_check_array_type(rb_funcall(num, id_divmod, 1, INT2FIX(1)));
+            if (NIL_P(ary)) {
+                goto typeerror;
+            }
+            i = rb_ary_entry(ary, 0);
+            f = rb_ary_entry(ary, 1);
+            t.tv_sec = NUM2LONG(i);
+            if (interval && t.tv_sec < 0)
+                rb_raise(rb_eArgError, "%s must be positive", tstr);
+            f = rb_funcall(f, id_mul, 1, INT2FIX(1000000000));
+            t.tv_nsec = NUM2LONG(f);
+        }
+        else {
+typeerror:
             rb_raise(rb_eTypeError, "can't convert %s into %s",
                      rb_obj_classname(num), tstr);
         }
-        i = rb_ary_entry(ary, 0);
-        f = rb_ary_entry(ary, 1);
-        t.tv_sec = NUM2LONG(i);
-	if (interval && t.tv_sec < 0)
-	    rb_raise(rb_eArgError, "%s must be positive", tstr);
-        f = rb_funcall(f, id_mul, 1, INT2FIX(1000000000));
-        t.tv_nsec = NUM2LONG(f);
 	break;
     }
     return t;
@@ -326,7 +337,7 @@ time_s_at(int argc, VALUE *argv, VALUE klass)
     return t;
 }
 
-static const char *months[] = {
+static const char months[][4] = {
     "jan", "feb", "mar", "apr", "may", "jun",
     "jul", "aug", "sep", "oct", "nov", "dec",
 };
@@ -349,7 +360,7 @@ obj2nsec(VALUE obj, long *nsec)
     if (TYPE(obj) == T_STRING) {
 	obj = rb_str_to_inum(obj, 10, Qfalse);
         *nsec = 0;
-        return NUM2LONG(obj) * 1000;
+        return NUM2LONG(obj);
     }
 
     ts = time_timespec(obj, 1);
@@ -483,7 +494,7 @@ leap_year_p(long y)
 static time_t
 timegm_noleapsecond(struct tm *tm)
 {
-    static int common_year_yday_offset[] = {
+    static const int common_year_yday_offset[] = {
 	-1,
 	-1 + 31,
 	-1 + 31 + 28,
@@ -498,7 +509,7 @@ timegm_noleapsecond(struct tm *tm)
 	-1 + 31 + 28 + 31 + 30 + 31 + 30 + 31 + 31 + 30 + 31 + 30
 	  /* 1    2    3    4    5    6    7    8    9    10   11 */
     };
-    static int leap_year_yday_offset[] = {
+    static const int leap_year_yday_offset[] = {
 	-1,
 	-1 + 31,
 	-1 + 31 + 29,
@@ -822,7 +833,10 @@ static time_t
 make_time_t(struct tm *tptr, int utc_p)
 {
     time_t t;
-    struct tm *tmp, buf;
+#ifdef NEGATIVE_TIME_T
+    struct tm *tmp;
+#endif
+    struct tm buf;
     buf = *tptr;
     if (utc_p) {
 #if defined(HAVE_TIMEGM)
@@ -1830,7 +1844,7 @@ time_yday(VALUE time)
  *  Returns <code>true</code> if <i>time</i> occurs during Daylight
  *  Saving Time in its time zone.
  *     
- *   CST6CDT:
+ *   # CST6CDT:
  *     Time.local(2000, 1, 1).zone    #=> "CST"
  *     Time.local(2000, 1, 1).isdst   #=> false
  *     Time.local(2000, 1, 1).dst?    #=> false
@@ -1838,7 +1852,7 @@ time_yday(VALUE time)
  *     Time.local(2000, 7, 1).isdst   #=> true
  *     Time.local(2000, 7, 1).dst?    #=> true
  *
- *   Asia/Tokyo:
+ *   # Asia/Tokyo:
  *     Time.local(2000, 1, 1).zone    #=> "JST"
  *     Time.local(2000, 1, 1).isdst   #=> false
  *     Time.local(2000, 1, 1).dst?    #=> false
@@ -2003,8 +2017,9 @@ rb_strftime(char **buf, const char *format, struct tm *time)
     if (flen == 0) {
 	return 0;
     }
+    errno = 0;
     len = strftime(*buf, SMALLBUF, format, time);
-    if (len != 0 || **buf == '\0') return len;
+    if (len != 0 || (**buf == '\0' && errno != ERANGE)) return len;
     for (size=1024; ; size*=2) {
 	*buf = xmalloc(size);
 	(*buf)[0] = '\0';
@@ -2017,7 +2032,7 @@ rb_strftime(char **buf, const char *format, struct tm *time)
 	 * format string, it's not failing for lack of room.
 	 */
 	if (len > 0 || size >= 1024 * flen) return len;
-	free(*buf);
+	xfree(*buf);
     }
     /* not reached */
 }
@@ -2097,7 +2112,7 @@ time_strftime(VALUE time, VALUE format)
 	    rb_str_cat(str, buf, len);
 	    p += strlen(p);
 	    if (buf != buffer) {
-		free(buf);
+		xfree(buf);
 		buf = buffer;
 	    }
 	    for (fmt = p; p < pe && !*p; ++p);
@@ -2109,7 +2124,7 @@ time_strftime(VALUE time, VALUE format)
 	len = rb_strftime(&buf, RSTRING_PTR(format), &tobj->tm);
     }
     str = rb_str_new(buf, len);
-    if (buf != buffer) free(buf);
+    if (buf != buffer) xfree(buf);
     rb_enc_copy(str, format);
     return str;
 }
@@ -2327,6 +2342,8 @@ time_load(VALUE klass, VALUE str)
 void
 Init_Time(void)
 {
+#undef rb_intern
+
     id_divmod = rb_intern("divmod");
     id_mul = rb_intern("*");
     id_submicro = rb_intern("submicro");
