@@ -21,6 +21,7 @@
 #include <stdlib.h>
 #include <errno.h>
 #include <assert.h>
+#include <ctype.h>
 
 #include <windows.h>
 #include <winbase.h>
@@ -66,7 +67,6 @@
 
 static struct ChildRecord *CreateChild(const char *, const char *, SECURITY_ATTRIBUTES *, HANDLE, HANDLE, HANDLE);
 static int has_redirection(const char *);
-static void StartSockets(void);
 int rb_w32_wait_events(HANDLE *events, int num, DWORD timeout);
 #if !defined(_WIN32_WCE)
 static int rb_w32_open_osfhandle(intptr_t osfhandle, int flags);
@@ -155,6 +155,15 @@ static struct {
     {	ERROR_INFLOOP_IN_RELOC_CHAIN,	ENOEXEC		},
     {	ERROR_FILENAME_EXCED_RANGE,	ENOENT		},
     {	ERROR_NESTING_NOT_ALLOWED,	EAGAIN		},
+#ifndef ERROR_PIPE_LOCAL
+#define ERROR_PIPE_LOCAL	229L
+#endif
+    {	ERROR_PIPE_LOCAL,		EPIPE		},
+    {	ERROR_BAD_PIPE,			EPIPE		},
+    {	ERROR_PIPE_BUSY,		EAGAIN		},
+    {	ERROR_NO_DATA,			EPIPE		},
+    {	ERROR_PIPE_NOT_CONNECTED,	EPIPE		},
+    {	ERROR_OPERATION_ABORTED,	EINTR		},
     {	ERROR_NOT_ENOUGH_QUOTA,		ENOMEM		},
     {	WSAENAMETOOLONG,		ENAMETOOLONG	},
     {	WSAENOTEMPTY,			ENOTEMPTY	},
@@ -166,8 +175,8 @@ static struct {
     {	WSAEMFILE,			EMFILE		},
 };
 
-static int
-map_errno(DWORD winerr)
+int
+rb_w32_map_errno(DWORD winerr)
 {
     int i;
 
@@ -187,16 +196,17 @@ map_errno(DWORD winerr)
     return EINVAL;
 }
 
-static char *NTLoginName;
+#define map_errno rb_w32_map_errno
 
+static const char *NTLoginName;
+
+static OSVERSIONINFO osver;
 #ifdef WIN95
 static DWORD Win32System = (DWORD)-1;
 
 DWORD
 rb_w32_osid(void)
 {
-    static OSVERSIONINFO osver;
-
     if (osver.dwPlatformId != Win32System) {
 	memset(&osver, 0, sizeof(OSVERSIONINFO));
 	osver.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
@@ -206,6 +216,19 @@ rb_w32_osid(void)
     return (Win32System);
 }
 #endif
+static DWORD Win32Version = (DWORD)-1;
+
+static DWORD
+rb_w32_osver(void)
+{
+    if (osver.dwMajorVersion != Win32Version) {
+	memset(&osver, 0, sizeof(OSVERSIONINFO));
+	osver.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
+	GetVersionEx(&osver);
+	Win32Version = osver.dwMajorVersion;
+    }
+    return (Win32Version);
+}
 
 #define IsWinNT() rb_w32_iswinnt()
 #define IsWin95() rb_w32_iswin95()
@@ -374,11 +397,10 @@ init_env(void)
     }
 
     if (!GetEnvironmentVariable("USER", env, sizeof env)) {
-	if (GetEnvironmentVariable("USERNAME", env, sizeof env) ||
-	    GetUserName(env, (len = sizeof env, &len))) {
+	if (GetEnvironmentVariable("USERNAME", env, sizeof env)) {
 	    SetEnvironmentVariable("USER", env);
 	}
-	else {
+	else if (!GetUserName(env, (len = sizeof env, &len))) {
 	    NTLoginName = "<Unknown>";
 	    return;
 	}
@@ -386,6 +408,18 @@ init_env(void)
     NTLoginName = strdup(env);
 }
 
+
+typedef BOOL (WINAPI *cancel_io_t)(HANDLE);
+static cancel_io_t cancel_io = NULL;
+
+static void
+init_func(void)
+{
+    if (!cancel_io)
+	cancel_io = (cancel_io_t)GetProcAddress(GetModuleHandle("kernel32"),
+						"CancelIo");
+}
+
 static void init_stdhandle(void);
 
 #if _MSC_VER >= 1400
@@ -396,14 +430,18 @@ static void invalid_parameter(const wchar_t *expr, const wchar_t *func, const wc
 #endif
 
 static CRITICAL_SECTION select_mutex;
-static BOOL fWinsock;
+static int NtSocketsInitialized = 0;
+static st_table *socklist = NULL;
 static char *envarea;
+
 static void
 exit_handler(void)
 {
-    if (fWinsock) {
+    if (NtSocketsInitialized) {
 	WSACleanup();
-	fWinsock = FALSE;
+	st_free_table(socklist);
+	socklist = NULL;
+	NtSocketsInitialized = 0;
     }
     if (envarea) {
 	FreeEnvironmentStrings(envarea);
@@ -412,16 +450,33 @@ exit_handler(void)
     DeleteCriticalSection(&select_mutex);
 }
 
+static void
+StartSockets(void)
+{
+    WORD version;
+    WSADATA retdata;
+
+    //
+    // initalize the winsock interface and insure that it's
+    // cleaned up at exit.
+    //
+    version = MAKEWORD(2, 0);
+    if (WSAStartup(version, &retdata))
+	rb_fatal ("Unable to locate winsock library!\n");
+    if (LOBYTE(retdata.wVersion) != 2)
+	rb_fatal("could not find version 2 of winsock dll\n");
+
+    socklist = st_init_numtable();
+
+    NtSocketsInitialized = 1;
+}
+
 //
 // Initialization stuff
 //
 void
 rb_w32_sysinit(int *argc, char ***argv)
 {
-
-    WORD version;
-    int ret;
-
 #if _MSC_VER >= 1400
     static void set_pioinfo_extra(void);
 
@@ -442,6 +497,8 @@ rb_w32_sysinit(int *argc, char ***argv)
 
     init_env();
 
+    init_func();
+
     init_stdhandle();
 
     InitializeCriticalSection(&select_mutex);
@@ -460,7 +517,7 @@ rb_w32_sysinit(int *argc, char ***argv)
 char *
 getlogin(void)
 {
-    return NTLoginName;
+    return (char *)NTLoginName;
 }
 
 #define MAXCHILDNUM 256	/* max num of child processes */
@@ -475,15 +532,6 @@ static struct ChildRecord {
     for (v = ChildRecord; v < ChildRecord + sizeof(ChildRecord) / sizeof(ChildRecord[0]); ++v)
 #define END_FOREACH_CHILD } while (0)
 
-static struct ChildRecord *
-FindFirstChildSlot(void)
-{
-    FOREACH_CHILD(child) {
-	if (child->pid) return child;
-    } END_FOREACH_CHILD;
-    return NULL;
-}
-
 static struct ChildRecord *
 FindChildSlot(rb_pid_t pid)
 {
@@ -703,10 +751,9 @@ rb_w32_join_argv(char *cmd, char *const *argv)
     return cmd;
 }
 
-static int socketpair_internal(int af, int type, int protocol, SOCKET *sv);
-
 rb_pid_t
-rb_w32_pipe_exec(const char *cmd, const char *prog, int mode, int *pipe, int *write_pipe)
+rb_w32_pipe_exec(const char *cmd, const char *prog, int mode, int *pipe,
+		 int *write_pipe)
 {
     struct ChildRecord* child;
     HANDLE hIn, hOut;
@@ -714,9 +761,6 @@ rb_w32_pipe_exec(const char *cmd, const char *prog, int mode, int *pipe, int *wr
     HANDLE hCurProc;
     SECURITY_ATTRIBUTES sa;
     BOOL reading, writing;
-    SOCKET pair[2];
-    int fd;
-    int binmode;
     int ret;
 
     /* Figure out what we're doing... */
@@ -731,7 +775,9 @@ rb_w32_pipe_exec(const char *cmd, const char *prog, int mode, int *pipe, int *wr
 	reading = TRUE;
 	writing = FALSE;
     }
-    binmode |= (mode & O_BINARY) ? O_BINARY : O_TEXT;
+    mode &= ~(O_RDWR|O_RDONLY|O_WRONLY);
+    if (!(mode & O_BINARY))
+	mode |= O_TEXT;
 
     sa.nLength              = sizeof (SECURITY_ATTRIBUTES);
     sa.lpSecurityDescriptor = NULL;
@@ -791,21 +837,31 @@ rb_w32_pipe_exec(const char *cmd, const char *prog, int mode, int *pipe, int *wr
 
 	/* associate handle to file descritor */
 	if (reading) {
-	    *pipe = rb_w32_open_osfhandle((intptr_t)hDupIn, _O_RDONLY | binmode);
+	    *pipe = rb_w32_open_osfhandle((intptr_t)hDupIn, O_RDONLY | mode);
 	    if (writing)
-		*write_pipe = rb_w32_open_osfhandle((intptr_t)hDupOut, _O_WRONLY | binmode);
+		*write_pipe = rb_w32_open_osfhandle((intptr_t)hDupOut,
+						    O_WRONLY | mode);
 	}
 	else {
-	    *pipe = rb_w32_open_osfhandle((intptr_t)hDupOut, _O_WRONLY | binmode);
+	    *pipe = rb_w32_open_osfhandle((intptr_t)hDupOut, O_WRONLY | mode);
 	}
 	if (hIn)
 	    CloseHandle(hIn);
 	if (hOut)
 	    CloseHandle(hOut);
-	if (*pipe == -1) {
-	    if (hDupIn)
+	if (reading && writing && *write_pipe == -1) {
+	    if (*pipe != -1)
+		rb_w32_close(*pipe);
+	    else
 		CloseHandle(hDupIn);
-	    if (hDupOut)
+	    CloseHandle(hDupOut);
+	    CloseChildHandle(child);
+	    break;
+	}
+	else if (*pipe == -1) {
+	    if (reading)
+		CloseHandle(hDupIn);
+	    else
 		CloseHandle(hDupOut);
 	    CloseChildHandle(child);
 	    break;
@@ -824,7 +880,6 @@ rb_w32_spawn(int mode, const char *cmd, const char *prog)
     DWORD exitcode;
 
     switch (mode) {
-      case P_WAIT:
       case P_NOWAIT:
       case P_OVERLAY:
 	break;
@@ -839,9 +894,6 @@ rb_w32_spawn(int mode, const char *cmd, const char *prog)
     }
 
     switch (mode) {
-      case P_WAIT:
-	rb_syswait(child->pid);
-	return NUM2INT(rb_last_status_get());
       case P_NOWAIT:
 	return child->pid;
       case P_OVERLAY:
@@ -1036,7 +1088,7 @@ typedef struct _NtCmdLineElement {
 #define NTSTRING 0x4	// element contains a quoted string
 
 static int
-insert(const char *path, VALUE vinfo)
+insert(const char *path, VALUE vinfo, void *enc)
 {
     NtCmdLineElement *tmpcurr;
     NtCmdLineElement ***tail = (NtCmdLineElement ***)vinfo;
@@ -1144,7 +1196,7 @@ skipspace(char *ptr)
 int 
 rb_w32_cmdvector(const char *cmd, char ***vec)
 {
-    int cmdlen, globbing, len, i;
+    int globbing, len;
     int elements, strsz, done;
     int slashes, escape;
     char *ptr, *base, *buffer, *cmdline;
@@ -1781,27 +1833,10 @@ rb_w32_open_osfhandle(intptr_t osfhandle, int flags)
 static int
 is_socket(SOCKET sock)
 {
-    char sockbuf[80];
-    int optlen;
-    int retval;
-    int result = TRUE;
-
-    optlen = sizeof(sockbuf);
-    RUBY_CRITICAL({
-	retval = getsockopt(sock, SOL_SOCKET, SO_TYPE, sockbuf, &optlen);
-	if (retval == SOCKET_ERROR) {
-	    int iRet;
-	    iRet = WSAGetLastError();
-	    if (iRet == WSAENOTSOCK || iRet == WSANOTINITIALISED)
-		result = FALSE;
-	}
-    });
-
-    //
-    // If we get here, then sock is actually a socket.
-    //
-
-    return result;
+    if (st_lookup(socklist, (st_data_t)sock, NULL))
+	return TRUE;
+    else
+	return FALSE;
 }
 
 int
@@ -1978,8 +2013,6 @@ rb_w32_fdisset(int fd, fd_set *set)
 
 #undef select
 
-static int NtSocketsInitialized = 0;
-
 static int
 extract_fd(fd_set *dst, fd_set *src, int (*func)(SOCKET))
 {
@@ -2020,9 +2053,9 @@ is_pipe(SOCKET sock) /* DONT call this for SOCKET! it clains it is PIPE. */
 {
     int ret;
 
-    RUBY_CRITICAL(
-	ret = (GetFileType((HANDLE)sock) == FILE_TYPE_PIPE)
-    );
+    RUBY_CRITICAL({
+	ret = (GetFileType((HANDLE)sock) == FILE_TYPE_PIPE);
+    });
 
     return ret;
 }
@@ -2147,6 +2180,7 @@ rb_w32_select(int nfds, fd_set *rd, fd_set *wr, fd_set *ex,
     fd_set cons_rd;
     fd_set else_rd;
     fd_set else_wr;
+    fd_set except;
     int nonsock = 0;
 
     if (nfds < 0 || (timeout && (timeout->tv_sec < 0 || timeout->tv_usec < 0))) {
@@ -2175,6 +2209,9 @@ rb_w32_select(int nfds, fd_set *rd, fd_set *wr, fd_set *ex,
     else_wr.fd_count = 0;
     nonsock += extract_fd(&else_wr, wr, is_not_socket);
 
+    except.fd_count = 0;
+    extract_fd(&except, ex, is_not_socket); // drop only
+
     r = 0;
     if (rd && rd->fd_count > r) r = rd->fd_count;
     if (wr && wr->fd_count > r) r = wr->fd_count;
@@ -2228,27 +2265,6 @@ rb_w32_select(int nfds, fd_set *rd, fd_set *wr, fd_set *ex,
     return r;
 }
 
-static void
-StartSockets(void)
-{
-    WORD version;
-    WSADATA retdata;
-
-    //
-    // initalize the winsock interface and insure that it's
-    // cleaned up at exit.
-    //
-    version = MAKEWORD(2, 0);
-    if (WSAStartup(version, &retdata))
-	rb_fatal ("Unable to locate winsock library!\n");
-    if (LOBYTE(retdata.wVersion) != 2)
-	rb_fatal("could not find version 2 of winsock dll\n");
-
-    fWinsock = TRUE;
-
-    NtSocketsInitialized = 1;
-}
-
 #undef accept
 
 int WSAAPI
@@ -2266,7 +2282,11 @@ rb_w32_accept(int s, struct sockaddr *addr, int *addrlen)
 	    s = -1;
 	}
 	else {
-	    s = rb_w32_open_osfhandle(r, O_RDWR|O_BINARY);
+	    s = rb_w32_open_osfhandle(r, O_RDWR|O_BINARY|O_NOINHERIT);
+	    if (s != -1)
+		st_insert(socklist, (st_data_t)r, (st_data_t)0);
+	    else
+		closesocket(r);
 	}
     });
     return s;
@@ -2400,75 +2420,131 @@ rb_w32_listen(int s, int backlog)
 }
 
 #undef recv
-
-int WSAAPI
-rb_w32_recv(int s, char *buf, int len, int flags)
-{
-    int r;
-    if (!NtSocketsInitialized) {
-	StartSockets();
-    }
-    RUBY_CRITICAL({
-	r = recv(TO_SOCKET(s), buf, len, flags);
-	if (r == SOCKET_ERROR)
-	    errno = map_errno(WSAGetLastError());
-    });
-    return r;
-}
-
 #undef recvfrom
-
-int WSAAPI
-rb_w32_recvfrom(int s, char *buf, int len, int flags,
-		 struct sockaddr *from, int *fromlen)
-{
-    int r;
-    if (!NtSocketsInitialized) {
-	StartSockets();
-    }
-    RUBY_CRITICAL({
-	r = recvfrom(TO_SOCKET(s), buf, len, flags, from, fromlen);
-	if (r == SOCKET_ERROR)
-	    errno = map_errno(WSAGetLastError());
-    });
-    return r;
-}
-
 #undef send
-
-int WSAAPI
-rb_w32_send(int s, const char *buf, int len, int flags)
-{
-    int r;
-    if (!NtSocketsInitialized) {
-	StartSockets();
-    }
-    RUBY_CRITICAL({
-	r = send(TO_SOCKET(s), buf, len, flags);
-	if (r == SOCKET_ERROR)
-	    errno = map_errno(WSAGetLastError());
-    });
-    return r;
-}
-
 #undef sendto
 
-int WSAAPI
-rb_w32_sendto(int s, const char *buf, int len, int flags, 
-	      const struct sockaddr *to, int tolen)
+static int
+overlapped_socket_io(BOOL input, int fd, char *buf, int len, int flags,
+		     struct sockaddr *addr, int *addrlen)
 {
     int r;
-    if (!NtSocketsInitialized) {
+    int ret;
+    int mode;
+    int flg;
+    WSAOVERLAPPED wol;
+    WSABUF wbuf;
+    int err;
+    SOCKET s;
+
+    if (!NtSocketsInitialized)
 	StartSockets();
+
+    s = TO_SOCKET(fd);
+    st_lookup(socklist, (st_data_t)s, (st_data_t *)&mode);
+    if (!cancel_io || (mode & O_NONBLOCK)) {
+	RUBY_CRITICAL({
+	    if (input) {
+		if (addr && addrlen)
+		    r = recvfrom(s, buf, len, flags, addr, addrlen);
+		else
+		    r = recv(s, buf, len, flags);
+	    }
+	    else {
+		if (addr && addrlen)
+		    r = sendto(s, buf, len, flags, addr, *addrlen);
+		else
+		    r = send(s, buf, len, flags);
+	    }
+	    if (r == SOCKET_ERROR)
+		errno = map_errno(WSAGetLastError());
+	});
     }
-    RUBY_CRITICAL({
-	r = sendto(TO_SOCKET(s), buf, len, flags, to, tolen);
-	if (r == SOCKET_ERROR)
-	    errno = map_errno(WSAGetLastError());
-    });
+    else {
+	DWORD size;
+	wbuf.len = len;
+	wbuf.buf = buf;
+	memset(&wol, 0, sizeof(wol));
+	RUBY_CRITICAL({
+	    wol.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
+	    if (input) {
+		flg = flags;
+		if (addr && addrlen)
+		    ret = WSARecvFrom(s, &wbuf, 1, &size, &flg, addr, addrlen,
+				      &wol, NULL);
+		else
+		    ret = WSARecv(s, &wbuf, 1, &size, &flg, &wol, NULL);
+	    }
+	    else {
+		if (addr && addrlen)
+		    ret = WSASendTo(s, &wbuf, 1, &size, flags, addr, *addrlen,
+				    &wol, NULL);
+		else
+		    ret = WSASend(s, &wbuf, 1, &size, flags, &wol, NULL);
+	    }
+	});
+
+	if (ret != SOCKET_ERROR) {
+	    r = size;
+	}
+	else if ((err = WSAGetLastError()) == WSA_IO_PENDING) {
+	    switch (rb_w32_wait_events_blocking(&wol.hEvent, 1, INFINITE)) {
+	      case WAIT_OBJECT_0:
+		RUBY_CRITICAL(
+		    ret = WSAGetOverlappedResult(s, &wol, &size, TRUE, &flg)
+		    );
+		if (ret) {
+		    r = size;
+		    break;
+		}
+		/* thru */
+	      default:
+		errno = map_errno(err);
+		/* thru */
+	      case WAIT_OBJECT_0 + 1:
+		/* interrupted */
+		r = -1;
+		cancel_io((HANDLE)s);
+		break;
+	    }
+	}
+	else {
+	    errno = map_errno(err);
+	    r = -1;
+	}
+	CloseHandle(&wol.hEvent);
+    }
+
     return r;
 }
 
+int WSAAPI
+rb_w32_recv(int fd, char *buf, int len, int flags)
+{
+    return overlapped_socket_io(TRUE, fd, buf, len, flags, NULL, NULL);
+}
+
+int WSAAPI
+rb_w32_recvfrom(int fd, char *buf, int len, int flags,
+		struct sockaddr *from, int *fromlen)
+{
+    return overlapped_socket_io(TRUE, fd, buf, len, flags, from, fromlen);
+}
+
+int WSAAPI
+rb_w32_send(int fd, const char *buf, int len, int flags)
+{
+    return overlapped_socket_io(FALSE, fd, (char *)buf, len, flags, NULL, NULL);
+}
+
+int WSAAPI
+rb_w32_sendto(int fd, const char *buf, int len, int flags, 
+	      const struct sockaddr *to, int tolen)
+{
+    return overlapped_socket_io(FALSE, fd, (char *)buf, len, flags,
+				(struct sockaddr *)to, &tolen);
+}
+
 #undef setsockopt
 
 int WSAAPI
@@ -2535,7 +2611,8 @@ open_ifs_socket(int af, int type, int protocol)
 		    if ((proto_buffers[i].dwServiceFlags1 & XP1_IFS_HANDLES) == 0)
 			continue;
 
-		    out = WSASocket(af, type, protocol, &(proto_buffers[i]), 0, 0);
+		    out = WSASocket(af, type, protocol, &(proto_buffers[i]), 0,
+				    WSA_FLAG_OVERLAPPED);
 		    break;
 		}
 	    }
@@ -2565,7 +2642,11 @@ rb_w32_socket(int af, int type, int protocol)
 	    fd = -1;
 	}
 	else {
-	    fd = rb_w32_open_osfhandle(s, O_RDWR|O_BINARY);
+	    fd = rb_w32_open_osfhandle(s, O_RDWR|O_BINARY|O_NOINHERIT);
+	    if (fd != -1)
+		st_insert(socklist, (st_data_t)s, (st_data_t)0);
+	    else
+		closesocket(s);
 	}
     });
     return fd;
@@ -2693,7 +2774,6 @@ rb_w32_getservbyport(int port, const char *proto)
 static int
 socketpair_internal(int af, int type, int protocol, SOCKET *sv)
 {
-    const char *localhost = NULL;
     SOCKET svr = INVALID_SOCKET, r = INVALID_SOCKET, w = INVALID_SOCKET;
     struct sockaddr_in sock_in4;
 #ifdef INET6
@@ -2786,8 +2866,20 @@ rb_w32_socketpair(int af, int type, int protocol, int *sv)
 
     if (socketpair_internal(af, type, protocol, pair) < 0)
 	return -1;
-    sv[0] = rb_w32_open_osfhandle(pair[0], O_RDWR|O_BINARY);
-    sv[1] = rb_w32_open_osfhandle(pair[1], O_RDWR|O_BINARY);
+    sv[0] = rb_w32_open_osfhandle(pair[0], O_RDWR|O_BINARY|O_NOINHERIT);
+    if (sv[0] == -1) {
+	closesocket(pair[0]);
+	closesocket(pair[1]);
+	return -1;
+    }
+    sv[1] = rb_w32_open_osfhandle(pair[1], O_RDWR|O_BINARY|O_NOINHERIT);
+    if (sv[1] == -1) {
+	rb_w32_close(sv[0]);
+	closesocket(pair[1]);
+	return -1;
+    }
+    st_insert(socklist, (st_data_t)pair[0], (st_data_t)0);
+    st_insert(socklist, (st_data_t)pair[1], (st_data_t)0);
 
     return 0;
 }
@@ -2826,6 +2918,7 @@ fcntl(int fd, int cmd, ...)
     va_list va;
     int arg;
     int ret;
+    int flag = 0;
     u_long ioctlArg;
 
     if (!is_socket(sock)) {
@@ -2840,17 +2933,21 @@ fcntl(int fd, int cmd, ...)
     va_start(va, cmd);
     arg = va_arg(va, int);
     va_end(va);
+    st_lookup(socklist, (st_data_t)sock, (st_data_t*)&flag);
     if (arg & O_NONBLOCK) {
+	flag |= O_NONBLOCK;
 	ioctlArg = 1;
     }
     else {
+	flag &= ~O_NONBLOCK;
 	ioctlArg = 0;
     }
     RUBY_CRITICAL({
 	ret = ioctlsocket(sock, FIONBIO, &ioctlArg);
-	if (ret == -1) {
+	if (ret == 0)
+	    st_insert(socklist, (st_data_t)sock, (st_data_t)flag);
+	else
 	    errno = map_errno(WSAGetLastError());
-	}
     });
 
     return ret;
@@ -2871,8 +2968,12 @@ poll_child_status(struct ChildRecord *child, int *stat_loc)
 	err = GetLastError();
 	if (err == ERROR_INVALID_PARAMETER)
 	    errno = ECHILD;
-	else
-	    errno = map_errno(GetLastError());
+	else {
+	    if (GetLastError() == ERROR_INVALID_HANDLE)
+		errno = EINVAL;
+	    else
+		errno = map_errno(GetLastError());
+	}
 	CloseChildHandle(child);
 	return -1;
     }
@@ -3760,7 +3861,7 @@ rb_w32_asynchronize(asynchronous_func_t func, uintptr_t self,
     });
 
     if (!thr) {
-	rb_fatal("failed to launch waiter thread:%d", GetLastError());
+	rb_fatal("failed to launch waiter thread:%ld", GetLastError());
     }
 
     return val;
@@ -3789,7 +3890,6 @@ rb_w32_get_environ(void)
     myenvtop = (char **)malloc(sizeof(char *) * (num + 1));
     for (env = envtop, myenv = myenvtop; *env; env += strlen(env) + 1) {
 	if (*env != '=') {
-	    int len = strlen(env) + 1;
 	    if (!(*myenv = strdup(env))) {
 		break;
 	    }
@@ -3824,6 +3924,40 @@ rb_w32_getpid(void)
     return pid;
 }
 
+
+rb_pid_t
+rb_w32_getppid(void)
+{
+    static long (WINAPI *pNtQueryInformationProcess)(HANDLE, int, void *, ULONG, ULONG *) = NULL;
+    rb_pid_t ppid = 0;
+
+    if (!IsWin95() && rb_w32_osver() >= 5) {
+	if (!pNtQueryInformationProcess) {
+	    HANDLE hNtDll = GetModuleHandle("ntdll.dll");
+	    if (hNtDll) {
+		pNtQueryInformationProcess = (long (WINAPI *)(HANDLE, int, void *, ULONG, ULONG *))GetProcAddress(hNtDll, "NtQueryInformationProcess");
+		if (pNtQueryInformationProcess) {
+		    struct {
+			long ExitStatus;
+			void* PebBaseAddress;
+			ULONG AffinityMask;
+			ULONG BasePriority;
+			ULONG UniqueProcessId;
+			ULONG ParentProcessId;
+		    } pbi;
+		    ULONG len;
+		    long ret = pNtQueryInformationProcess(GetCurrentProcess(), 0, &pbi, sizeof(pbi), &len);
+		    if (!ret) {
+			ppid = pbi.ParentProcessId;
+		    }
+		}
+	    }
+	}
+    }
+
+    return ppid;
+}
+
 int
 rb_w32_fclose(FILE *fp)
 {
@@ -3857,6 +3991,7 @@ rb_w32_close(int fd)
 	return _close(fd);
     }
     _set_osfhnd(fd, (SOCKET)INVALID_HANDLE_VALUE);
+    st_delete(socklist, (st_data_t *)&sock, NULL);
     _close(fd);
     errno = save_errno;
     if (closesocket(sock) == SOCKET_ERROR) {
@@ -3884,8 +4019,12 @@ rb_w32_write(int fd, const void *buf, size_t size)
 {
     SOCKET sock = TO_SOCKET(fd);
 
-    if (!is_socket(sock))
-	return write(fd, buf, size);
+    if (!is_socket(sock)) {
+	size_t ret = write(fd, buf, size);
+	if ((int)ret < 0 && errno == EINVAL)
+	    errno = map_errno(GetLastError());
+	return ret;
+    }
     else
 	return rb_w32_send(fd, buf, size, 0);
 }
@@ -3918,9 +4057,7 @@ int
 rb_w32_utime(const char *path, const struct utimbuf *times)
 {
     HANDLE hFile;
-    SYSTEMTIME st;
     FILETIME atime, mtime;
-    struct tm *tm;
     struct stati64 stat;
     int ret = 0;
 
